/******************************** CPPFile *****************************

* FileName [CEDagVerifier.cpp]

* PackageName [main]

* Synopsis [Method definitions of CEDagVerifier class.]

* SeeAlso [CEDagVerifier.h]

* Author [Sagar Chaki]

* Copyright [ Copyright (c) 2002 by Carnegie Mellon University. All
* Rights Reserved. This software is for educational purposes only.
* Permission is given to academic institutions to use, copy, and
* modify this software and its documentation provided that this
* introductory message is not removed, that this software and its
* documentation is used for the institutions' internal research and
* educational purposes, and that no monies are exchanged. No guarantee
* is expressed or implied by the distribution of this code. Send
* bug-reports and/or questions to: chaki+@cs.cmu.edu. ]

**********************************************************************/

#include <cstdio>
#include <cassert>
#include <string>
#include <map>
#include <set>
#include <list>
#include <vector>
#include <stack>
#include <bitset>
using namespace std;

#include "Util.h"
#include "Node.h"
#include "LtsInfo.h"
#include "Action.h"
#include "ActionManager.h"
#include "ProcInfo.h"
#include "Database.h"
#include "Predicate.h"
#include "PredSet.h"
#include "ContLoc.h"
#include "ImplState.h"
#include "LocalAbsLts.h"
#include "GlobalAbsLts.h"
#include "GlobalCEDag.h"
#include "ProcManager.h"
#include "ConcCEDag.h"
#include "CEDagVerifier.h"
#include "StdcParser.h"

/*********************************************************************/
//define static fields
/*********************************************************************/
const int CEDagVerifier::CE_VALID = 11600;
const int CEDagVerifier::CE_INVALID_DAG = 11610;
const int CEDagVerifier::CE_INVALID_PATH = 11620;

/*********************************************************************/
//verify the projection of the global CE Dag on this procedure. if the
//projection is invalid refine it and return CE_INVALID_DAG. otherwise
//check the paths in the projection. if some path is invalid add it to
//the set of spurious paths and return CE_INVALID_PATH. if every path
//is valid return CE_VALID.
/*********************************************************************/
int CEDagVerifier::Run()
{
  //if checking simulation of ERROR state reachability
  if((Database::CONF_TYPE == Database::CONF_SIMUL) || (Database::CONF_TYPE == Database::CONF_REACH)) {
    return RunFsp();
  }
  //if checking LTL formula
  else if(Database::CONF_TYPE == Database::CONF_LTL) {
    return RunLtl();
  }
  //impossible
  else assert(false);
}

/*********************************************************************/
//the top-level routine when we are checking for simulation or ERROR
//state reachability
/*********************************************************************/
int CEDagVerifier::RunFsp()
{
  //initialise the concrete ce dag
  ConcCEDag concDag;
  if(!concDag.ComputeLocLabelFsp(procManager)) {
    int ssn = Database::GetSpecStateNum();
    int locState = GlobalAbsLts::CompToAbsStates(GlobalCEDag::stateLabel[0] / ssn)[procManager.index];
    set<const ImplState*> initImpls;
    set<const ImplState*> is = procManager.initLoc->GetImplStates(procManager.guard);
    for(set<const ImplState*>::const_iterator i = is.begin();i != is.end();i++) {
      if((*i)->GetAbsId() == locState) initImpls.insert(*i);
    }
    assert(!initImpls.empty());
    procManager.absLtsRefine.clear();
    CreateSpuriousAbsLtsCEInfo(0,initImpls,concDag);
    return CE_INVALID_DAG;
  }

  //check out the concrete dag
  spuriousCE = false;
  if(Database::CHECK_CE_BRANCHING) VerifyConcreteDag(concDag);
  else {
    set<ConcCEDag> concPaths; concDag.DagToPaths(concPaths,procManager);
    for(set<ConcCEDag>::const_iterator i = concPaths.begin();i != concPaths.end();i++) {
      VerifyConcreteDag(*i);
    }
    if(!spuriousCE) VerifyConcreteDag(concDag);
  }
  return spuriousCE ? CE_INVALID_PATH : CE_VALID;
}

/*********************************************************************/
//update the pre or postcondition after a statement
/*********************************************************************/
void CEDagVerifier::UpdatePrePost(set<ContLoc*> &expLocs,set<Expr> &expSet,const int node,const ConcCEDag &concDag,const bool pre) const
{
  const Action &act = concDag.actLabel[node];
  if(procManager.absLts.GetActs().count(act) != 0) {
    const ContLoc *loc = concDag.locLabel[node];
    int locType = loc->GetLocType();
    //for return and final locations
    if((locType == ContLoc::RETURN) || (locType == ContLoc::FINAL)) {}
    //for assignment locations
    else if(locType == ContLoc::ASSIGN_EXPR) {
      expSet = pre ? UpdateWP(expSet,loc->GetLocLhs(),loc->GetLocRhs()) : 
	UpdateSP(expSet,loc->GetLocLhs(),loc->GetLocRhs());
      //check if the assignment corresponds to a predicate inferred as
      //a result of some branch X. in that case add X as a possible
      //explanation. update the cache if required.
      Expr cond = ExprManager::GetBinaryExpr(loc->GetLocLhs(),loc->GetLocRhs(),MAGIC_EQ_OP);
      if(procManager.assignToBranch.count(cond) == 0) {
	for(map< ContLoc*,map<ContLoc*,PredSet> >::const_iterator j = procManager.locToPreds.begin();j != procManager.locToPreds.end();j++) {
	  bool locFlag = true;
	  for(map<ContLoc*,PredSet>::const_iterator k = j->second.begin();locFlag && (k != j->second.end());k++) {
	    const vector<Predicate> &preds = k->second.GetPreds();
	    for(vector<Predicate>::const_iterator l = preds.begin();locFlag && (l != preds.end());l++) {
	      const vector<Expr> &vte = l->GetValToExpr();
	      for(vector<Expr>::const_iterator m = vte.begin();m != vte.end();m++) {
		if(ExprManager::EquivalentTo(cond,*m) || ExprManager::NegationOf(cond,*m)) { 
		  procManager.assignToBranch[cond].insert(j->first);
		  locFlag = false;
		  break;
		}
	      }
	    }
	  }
	}
      }
      expLocs.insert(procManager.assignToBranch[cond].begin(),procManager.assignToBranch[cond].end());
    }
    //for branch locations
    else if(locType == ContLoc::BRANCH) {
      for(set<int>::const_iterator j = concDag.trans[node].begin();j != concDag.trans[node].end();j++) {
	const ContLoc *succLoc = concDag.locLabel[*j];
	if((loc->GetTSkipped().empty() && (succLoc == loc->GetTSucc())) ||
	   (!loc->GetTSkipped().empty() && (succLoc == loc->GetTSkipped().front()))) {
	  expSet.insert(loc->GetLocExpr());
	} else if((loc->GetESkipped().empty() && (succLoc == loc->GetESucc())) ||
		  (!loc->GetESkipped().empty() && (succLoc == loc->GetESkipped().front()))) {
	  expSet.insert(ExprManager::NegateExpr(loc->GetLocExpr()));
	} else assert(false);
      }
      expLocs.insert((ContLoc*)(loc));
    }
    //for call locations
    else if(locType == ContLoc::ASSIGN_CALL) {
      assert(loc->GetCallSucc() == NULL);
      if(act.IsAssign()) {
	list<string> params = loc->GetProcInfo()->GetParams();
	list<Expr> elist;
	for(list<string>::const_iterator l = params.begin();l != params.end();l++) {
	  elist.push_back(ExprManager::GetIdExpr(*l)); 
	}
	pair<Expr,Expr> comps = act.GetConcreteAssignExprs(elist);
	expSet = pre ? UpdateWP(expSet,comps.first,comps.second) : 
	  UpdateSP(expSet,comps.first,comps.second);
      }
    }
    //impossible
    else assert(false);
  }
}

/*********************************************************************/
//check the validity of a concrete ce dag. if the dag is spurious add
//it to the global collection.
/*********************************************************************/
void CEDagVerifier::VerifyConcreteDag(const ConcCEDag &concDag)
{
  //the set of branch locations that can explain this path's
  //infeasibility
  set<ContLoc*> expLocs;
  
  //compute the weakest precondition at each node of the graph in
  //reverse topological order
  vector< set<Expr> > dagToWP(concDag.locLabel.size());
  vector<int> topo; concDag.ComputeTopo(topo);
  for(vector<int>::const_iterator i = topo.begin();i != topo.end();i++) {
    for(set<int>::const_iterator j = concDag.trans[*i].begin();j != concDag.trans[*i].end();j++) {
      dagToWP[*i].insert(dagToWP[*j].begin(),dagToWP[*j].end());
    }
    UpdatePrePost(expLocs,dagToWP[*i],*i,concDag,true);
  }
  
  //add the constraint for the initial state
  dagToWP[0].insert(procManager.guard);
  
  //check for spuriousness
  if(ExprManager::ProveImplies(dagToWP[0],set<Expr>())) {
    Util::Message(3,"This is a spurious path ....\n");
    spuriousCE = true;
    //add the path to the set of spurious paths and update the set of
    //maps if necessary
    if(procManager.dagToLocs.count(concDag) == 0) {
      //procManager.dagToLocs[dag] = procManager.branchLocs;
      procManager.dagToLocs[concDag] = expLocs;
    }
  } 
  //if this is a real counter example display it
  else concDag.Display(procManager);
}

/*********************************************************************/
//given a set of expressions and an assignment return the set of
//strongest postconditions of the first argument w.r.t. to the second
/*********************************************************************/
set<Expr> CEDagVerifier::UpdateSP(const set<Expr> &old,const Expr &lhs,const Expr &rhs) const
{
  //the skolem constant
  Expr skolem = ExprManager::GetIdExpr(Util::NewTempVar());

  set<Expr> res;
  for(set<Expr>::const_iterator i = old.begin();i != old.end();i++) {
    res.insert(ExprManager::ComputeWP(lhs,skolem,*i));
  }

  Expr x = ExprManager::ComputeWP(lhs,skolem,rhs);
  Expr y = ExprManager::GetBinaryExpr(lhs,x,MAGIC_EQ_OP);
  res.insert(y);

  return res;
}

/*********************************************************************/
//given a set of expressions and an assignment return the set of
//weakest preconditions of the first argument w.r.t. to the second
/*********************************************************************/
set<Expr> CEDagVerifier::UpdateWP(const set<Expr> &old,const Expr &lhs,const Expr &rhs) const
{
  set<Expr> res;
  for(set<Expr>::const_iterator i = old.begin();i != old.end();i++) {
    res.insert(ExprManager::ComputeWP(lhs,rhs,*i));
  }
  return res;
}

/*********************************************************************/
//create information that explains why the CE in the abstract LTS is
//spurious. this information can be used to refine the abstraction.
/*********************************************************************/
void CEDagVerifier::CreateSpuriousAbsLtsCEInfo(const int node,const set<const ImplState*> &impls,const ConcCEDag &concDag) const
{
  int ssn = Database::GetSpecStateNum();
  const Action &act = concDag.actLabel[node];
  if(procManager.absLts.GetActs().count(act) == 0) {
    for(set<int>::const_iterator i = concDag.trans[node].begin();i != concDag.trans[node].end();i++) {
      int locState = GlobalAbsLts::CompToAbsStates(GlobalCEDag::stateLabel[*i] / ssn)[procManager.index];
      set<const ImplState*> nextImpls;
      for(set<const ImplState*>::const_iterator j = impls.begin();j != impls.end();j++) {
	if((*j)->GetAbsId() == locState) nextImpls.insert(*j);
      }
      assert(!nextImpls.empty());
      CreateSpuriousAbsLtsCEInfo(*i,nextImpls,concDag);
    }
  } else {
    int locState = GlobalAbsLts::CompToAbsStates(GlobalCEDag::stateLabel[node] / ssn)[procManager.index];	
    set<const ImplState*> succImpls;
    for(set<const ImplState*>::const_iterator j = impls.begin();j != impls.end();j++) {
      set<const ImplState*> is = (*j)->GetSuccsOnAction(act);
      succImpls.insert(is.begin(),is.end());
    }
    if(succImpls.empty()) {
      procManager.absLtsRefine[locState].insert(act);
    } else {
      for(set<int>::const_iterator i = concDag.trans[node].begin();i != concDag.trans[node].end();i++) {
	int nextState = GlobalAbsLts::CompToAbsStates(GlobalCEDag::stateLabel[*i] / ssn)[procManager.index];
	set<const ImplState*> nextImpls;
	for(set<const ImplState*>::const_iterator j = succImpls.begin();j != succImpls.end();j++) {
	  if((*j)->GetAbsId() == nextState) nextImpls.insert(*j);
	}
	if(nextImpls.empty()) {
	  procManager.absLtsRefine[locState].insert(act);
	} else {
	  CreateSpuriousAbsLtsCEInfo(*i,nextImpls,concDag);
	}
      }
    }
  }
}

/*********************************************************************/
//the top-level routine when we are checking for LTL property
/*********************************************************************/
int CEDagVerifier::RunLtl()
{
  ConcCEDag concDag;
  if(!concDag.ComputeLocLabelLtl(procManager)) return CE_INVALID_DAG;

  set<ContLoc*> expLocs;
  set<Expr> expSet,loopInv;
  expSet.insert(procManager.guard);
  bool looped = false;
  size_t currNode = 0;
  while(true) {
    UpdatePrePost(expLocs,expSet,currNode,concDag,false);
    size_t nextNode = *(concDag.trans[currNode].begin());
    if(nextNode <= currNode) {
      if(looped) {
	if(ExprManager::ProveImplies(expSet,set<Expr>())) {
	  Util::Message(3,"This is a spurious path ....\n");
	  if(procManager.dagToLocs.count(concDag) == 0) {
	    //procManager.dagToLocs[dag] = procManager.branchLocs;
	    procManager.dagToLocs[concDag] = expLocs;
	  }
	  return CE_INVALID_PATH;
	} else if(ExprManager::ProveImplies(expSet,loopInv) && ExprManager::ProveImplies(loopInv,expSet)) {
	  concDag.Display(procManager);
	  return CE_VALID;
	}
      } else looped = true;
      loopInv.insert(expSet.begin(),expSet.end());
    }
    currNode = nextNode;
  }
  assert(false);
}

/*********************************************************************/
//end of CEDagVerifier.cpp
/*********************************************************************/
