/******************************** CPPFile *****************************

* FileName [ConcCEDag.cpp]

* PackageName [main]

* Synopsis [Method definitions of ConcCEDag class.]

* SeeAlso [ConcCEDag.h]

* Author [Sagar Chaki]

* Copyright [ Copyright (c) 2002 by Carnegie Mellon University. All
* Rights Reserved. This software is for educational purposes only.
* Permission is given to academic institutions to use, copy, and
* modify this software and its documentation provided that this
* introductory message is not removed, that this software and its
* documentation is used for the institutions' internal research and
* educational purposes, and that no monies are exchanged. No guarantee
* is expressed or implied by the distribution of this code. Send
* bug-reports and/or questions to: chaki+@cs.cmu.edu. ]

**********************************************************************/

#include <cassert>
#include <string>
#include <list>
#include <set>
#include <map>
#include <vector>
#include <bitset>
using namespace std;

#include "Util.h"
#include "Node.h"
#include "Action.h"
#include "Database.h"
#include "Predicate.h"
#include "PredSet.h"
#include "ContLoc.h"
#include "ImplState.h"
#include "LocalAbsLts.h"
#include "GlobalAbsLts.h"
#include "GlobalCEDag.h"
#include "ProcManager.h"
#include "ConcCEDag.h"
#include "CEDagVerifier.h"

/*********************************************************************/
//operators
/*********************************************************************/
const ConcCEDag &ConcCEDag::operator = (const ConcCEDag &rhs)
{
  locLabel = rhs.locLabel;
  actLabel = rhs.actLabel;
  trans = rhs.trans;
  return *this;
}

bool ConcCEDag::operator == (const ConcCEDag &rhs) const
{
  return (locLabel == rhs.locLabel);
}

bool ConcCEDag::operator < (const ConcCEDag &rhs) const
{
  return (locLabel < rhs.locLabel);
}

/*********************************************************************/
//compute reverse topological order
/*********************************************************************/
void ConcCEDag::ComputeTopo(vector<int> &topo) const
{
  set<int> visited;
  ComputeTopoWithRoot(0,visited,topo);
  assert(topo.size() == locLabel.size());
}

/*********************************************************************/
//compute reverse topological order of a sub-dag
/*********************************************************************/
void ConcCEDag::ComputeTopoWithRoot(const int node,set<int> &visited,vector<int> &topo) const
{
  visited.insert(node);
  for(set<int>::const_iterator i = trans[node].begin();i != trans[node].end();i++) {
    if(visited.count(*i) == 0) ComputeTopoWithRoot(*i,visited,topo);
  }
  topo.push_back(node);
}

/*********************************************************************/
//compute the control location label
/*********************************************************************/
bool ConcCEDag::ComputeLocLabelFsp(const ProcManager &procManager)
{
  locLabel.resize(GlobalCEDag::stateLabel.size());
  actLabel = GlobalCEDag::actLabel;
  trans = GlobalCEDag::trans;
  //if not using LTS abstraction
  if(!Database::USE_LTS_ABSTRACTION) {
    ComputeLocLabelNoLtsAbs(procManager);
    return true;
  }
  //compute the reverse topological order
  vector<int> topo; ComputeTopo(topo);
  //compute set of possible impl states
  vector<ImStSet> dagToImpls(GlobalCEDag::stateLabel.size());
  int ssn = Database::GetSpecStateNum();
  for(vector<int>::const_iterator i = topo.begin();i != topo.end();i++) {
    //intersect successors
    dagToImpls[*i] = ImStSet::GetComplete();
    for(set<int>::const_iterator j = trans[*i].begin();j != trans[*i].end();j++) {
      dagToImpls[*i].Intersect(dagToImpls[*j]);
    }
    //take pre-image if necessary
    const Action &act = actLabel[*i];
    if(procManager.absLts.GetActs().count(act) != 0) {
      dagToImpls[*i] = dagToImpls[*i].GetPredsOnAction(act,procManager);
    }
    //project on global ce dag
    int locState = GlobalAbsLts::CompToAbsStates(GlobalCEDag::stateLabel[*i] / ssn)[procManager.index];
    const set<const ImplState*> ele = dagToImpls[*i].GetEle(procManager);
    set<const ImplState*> res;
    for(set<const ImplState*>::const_iterator j = ele.begin();j != ele.end();j++) {
      if((*j)->GetAbsId() == locState) {
	res.insert(*j);
      }
    }
    dagToImpls[*i] = ImStSet(res,procManager);
  }
  
  //check if some initial impl state can represent the root
  ImStSet initImpls = ImStSet(procManager.initLoc->GetImplStates(procManager.guard),procManager);
  initImpls.Intersect(dagToImpls[0]);
  if(initImpls.Empty()) return false;

  //select a impl state and then cont loc for each node
  vector<const ImplState*> dagToImpl(GlobalCEDag::stateLabel.size());
  dagToImpl[0] = initImpls.Pick(procManager);
  for(vector<int>::reverse_iterator i = topo.rbegin();i != topo.rend();i++) {
    locLabel[*i] = dagToImpl[*i]->GetContLoc();
    const Action &act = actLabel[*i];
    set<const ImplState*> succs = dagToImpl[*i]->GetSuccsOnAction(act);
    for(set<int>::const_iterator j = trans[*i].begin();j != trans[*i].end();j++) {
      //if this actions is exported
      if(procManager.absLts.GetActs().count(act) != 0) {
	bool found = false;
	for(set<const ImplState*>::const_iterator k = succs.begin();k != succs.end();k++) {
	  if(dagToImpls[*j].Contains(*k)) {
	    dagToImpl[*j] = *k;
	    found = true;
	    break;
	  }
	}
	assert(found);
      }
      //if this action is not exported
      else {
	assert(dagToImpls[*j].Contains(dagToImpl[*i]));
	dagToImpl[*j] = dagToImpl[*i];
      }
    } 
  }
  //success
  return true;
}

/*********************************************************************/
//compute the loc label when no lts abstraction is used
/*********************************************************************/
void ConcCEDag::ComputeLocLabelNoLtsAbs(const ProcManager &procManager)
{
  //compute reverse topological order
  vector<int> topo; ComputeTopo(topo);
  //compute possible impl states
  map<int,const ImplState*> idToImpl;
  for(set<const ImplState*>::const_iterator i = procManager.reachImpl.begin();i != procManager.reachImpl.end();i++) {
    idToImpl[(*i)->GetAbsId()] = (*i);
  } 
  int ssn = Database::GetSpecStateNum();
  for(vector<int>::const_iterator i = topo.begin();i != topo.end();i++) {
    int locState = GlobalAbsLts::CompToAbsStates(GlobalCEDag::stateLabel[*i] / ssn)[procManager.index];
    locLabel[*i] = idToImpl[locState]->GetContLoc();
  }
}

/*********************************************************************/
//compute the control location label for ltl model checking. if no
//labeling exists, generate the info for refining lts abstraction.
/*********************************************************************/
bool ConcCEDag::ComputeLocLabelLtl(ProcManager &procManager)
{
  //if not using LTS abstraction
  if(!Database::USE_LTS_ABSTRACTION) {
    locLabel.resize(GlobalCEDag::stateLabel.size());
    actLabel = GlobalCEDag::actLabel;
    trans = GlobalCEDag::trans;
    ComputeLocLabelNoLtsAbs(procManager);
    return true;
  }

  //map from local states to nodes
  int ssn = Database::GetSpecStateNum();
  map< int,set<size_t> > stateToDag;
  for(size_t i = 0;i < GlobalCEDag::stateLabel.size();i++) {
    stateToDag[GlobalAbsLts::CompToAbsStates(GlobalCEDag::stateLabel[i] / ssn)[procManager.index]].insert(i);
  }

  //map from nodes to candidate impl state sets
  vector< set<const ImplState*> > dagToIS(GlobalCEDag::stateLabel.size());
  const set<const ImplState*> &ri = procManager.GetReachImpl();
  for(set<const ImplState*>::const_iterator i = ri.begin();i != ri.end();i++) {
    const set<size_t> &x = stateToDag[(*i)->GetAbsId()];
    for(set<size_t>::const_iterator j = x.begin();j != x.end();j++) {
      dagToIS[*j].insert(*i);
    }
  }

  //first simulate the loop backward
  size_t currNode = GlobalCEDag::stateLabel.size() - 1;
  size_t loopBegin = *(GlobalCEDag::trans[currNode].begin());
  vector<ImStSet> dagToImpls1(GlobalCEDag::stateLabel.size());
  dagToImpls1[loopBegin] = ImStSet(dagToIS[loopBegin],procManager);
  bool looped = false;
  bool valid = true;
  while(true) {
    size_t nextNode = *(GlobalCEDag::trans[currNode].begin());
    const Action &act = GlobalCEDag::actLabel[currNode];
    ImStSet currSt = dagToImpls1[nextNode];
    if(procManager.absLts.GetActs().count(act) != 0) {
      currSt = currSt.GetPredsOnAction(act,procManager);
    }
    currSt.Intersect(ImStSet(dagToIS[currNode],procManager));
    if(currSt.Empty()) {
      valid = false;
      break;
    }
    if(currNode == loopBegin) looped = true;
    if(looped && (currSt == dagToImpls1[currNode])) break;
    dagToImpls1[currNode] = currSt;
    currNode = (currNode == loopBegin) ? (GlobalCEDag::stateLabel.size() - 1) : (currNode - 1);
  }

  //if not already invalid
  if(valid) {
    vector<ImStSet> dagToImpls2(GlobalCEDag::stateLabel.size());
    for(int i = GlobalCEDag::stateLabel.size() - 1;i >= 0;i--) {
      const Action &act = GlobalCEDag::actLabel[i];     
      ImStSet currSt = (static_cast<size_t>(i) == (GlobalCEDag::stateLabel.size() - 1)) ? dagToImpls1[loopBegin] : dagToImpls2[i+1];
      if(procManager.absLts.GetActs().count(act) != 0) {
	currSt = currSt.GetPredsOnAction(act,procManager);
      }
      currSt.Intersect(ImStSet(dagToIS[i],procManager));
      if(i == 0) currSt.Intersect(ImStSet(procManager.initLoc->GetImplStates(procManager.guard),procManager));
      if(currSt.Empty()) {
	valid = false;
	break;
      }
      dagToImpls2[i] = currSt;
    }
    if(valid) {
      locLabel.clear();
      actLabel.clear();
      trans.clear();
      vector<const ImplState*> ivec;
      ivec.push_back(dagToImpls2[0].Pick(procManager));
      actLabel.push_back(GlobalCEDag::actLabel[0]);
      trans.push_back(set<int>());
      for(size_t i = 1;i < GlobalCEDag::stateLabel.size();i++) {
	const Action &act = actLabel.back();
	if(procManager.absLts.GetActs().count(act) == 0) {
	  ivec.push_back(ivec.back());
	} else {
	  ImStSet x(ivec.back()->GetSuccsOnAction(act),procManager);
	  x.Intersect(dagToImpls2[i]);
	  ivec.push_back(x.Pick(procManager));
	}
	actLabel.push_back(GlobalCEDag::actLabel[i]);
	trans.back().insert(i);
	trans.push_back(set<int>());
      }
      size_t loopLength = GlobalCEDag::stateLabel.size() - loopBegin;
      currNode = loopBegin;
      while(true) {
	const Action &act = actLabel.back();
	set<const ImplState*> succImpls;
	if(procManager.absLts.GetActs().count(act) == 0) {
	  succImpls.insert(ivec.back());
	} else {
	  succImpls = ivec.back()->GetSuccsOnAction(act);
	}
	ImStSet succSet(succImpls,procManager);
	if(currNode == loopBegin) {
	  looped = false;
	  for(size_t i = loopBegin;i < ivec.size();i += loopLength) {
	    if(succSet.Contains(ivec[i])) {
	      trans.back().clear();
	      trans.back().insert(i);
	      looped = true;
	    }
	  }
	  if(looped) break;
	}
	ivec.push_back(succSet.Pick(procManager));
	actLabel.push_back(GlobalCEDag::actLabel[currNode]);
	trans.back().insert(trans.size());
	trans.push_back(set<int>());
	currNode = (currNode == (GlobalCEDag::stateLabel.size() - 1)) ? loopBegin : (currNode + 1);
      }
      for(size_t i = 0;i < ivec.size();i++) {
	locLabel.push_back(ivec[i]->GetContLoc());
      }
      return true;
    }
  }
  
  //if invalid
  assert(!valid);
  for(size_t i = 0;i < GlobalCEDag::stateLabel.size();i++) {
    dagToImpls1[i] = ImStSet(dagToIS[i],procManager);
  }
  procManager.absLtsRefine.clear();
  currNode = 0;
  ImStSet currSt = dagToImpls1[currNode];
  currSt.Intersect(ImStSet(procManager.initLoc->GetImplStates(procManager.guard),procManager));
  while(true) {
    const Action &act = GlobalCEDag::actLabel[currNode];
    if(procManager.absLts.GetActs().count(act) != 0) {
      currSt = currSt.GetSuccsOnAction(act,procManager);
    }
    size_t nextNode = *(GlobalCEDag::trans[currNode].begin());
    currSt.Intersect(dagToImpls1[nextNode]);
    if(currSt.Empty()) {
      int locState = GlobalAbsLts::CompToAbsStates(GlobalCEDag::stateLabel[currNode] / ssn)[procManager.index];
      procManager.absLtsRefine[locState].insert(act);
      return false;
    }
    currNode = nextNode;
  }
  assert(false);
}

/*********************************************************************/
//compute the set of concrete paths in the ce dag
/*********************************************************************/
void ConcCEDag::DagToPaths(set<ConcCEDag> &concPaths,const ProcManager &procManager) const
{
  typedef pair<const ContLoc*,Action> ConcEle;
  typedef list<ConcEle> ConcPath;
  vector<int> topo; ComputeTopo(topo);
  vector< set<ConcPath> > dagToPaths(locLabel.size());
  for(vector<int>::const_iterator i = topo.begin();i != topo.end();i++) {
    for(set<int>::const_iterator j = trans[*i].begin();j != trans[*i].end();j++) {
      dagToPaths[*i].insert(dagToPaths[*j].begin(),dagToPaths[*j].end());
    }
    const Action &act = actLabel[*i];
    if(procManager.absLts.GetActs().count(act) != 0) {
      set<ConcPath> x = dagToPaths[*i];
      if(x.empty()) x.insert(ConcPath());
      dagToPaths[*i].clear();
      for(set<ConcPath>::iterator j = x.begin();j != x.end();j++) {
	ConcPath y = *j;
	y.push_front(pair<const ContLoc*,Action>(locLabel[*i],act));
	dagToPaths[*i].insert(y);
      }
    }
  }
  for(set<ConcPath>::const_iterator i = dagToPaths[0].begin();i != dagToPaths[0].end();i++) {
    ConcCEDag x;
    x.locLabel.resize(i->size());
    x.actLabel.resize(i->size());
    x.trans.resize(i->size());
    int count = 0;
    for(list<ConcEle>::const_iterator j = i->begin();j != i->end();j++,count++) {
      x.locLabel[count] = j->first;
      x.actLabel[count] = j->second;
      if(count > 0) x.trans[count - 1].insert(count);
    }    
    concPaths.insert(x);
  }
}

/*********************************************************************/
//display to standard output
/*********************************************************************/
void ConcCEDag::Display(const ProcManager &procManager) const
{
  list<const ContLoc*> locStack;
  locStack.push_back(locLabel[0]);
  list<Action> actStack;
  if((Database::CONF_TYPE == Database::CONF_SIMUL) || (Database::CONF_TYPE == Database::CONF_REACH)) {
    DisplayWithRoot(0,locStack,actStack,procManager);
  } else if(Database::CONF_TYPE == Database::CONF_LTL) {
    int i = 0;
    while(actStack.size() <= locLabel.size()) {
      actStack.push_back(actLabel[i]);
      assert(trans[i].size() == 1);
      i = *(trans[i].begin());
      locStack.push_back(locLabel[i]);
    }
    DisplayPath(locStack,actStack,procManager);
  } else assert(false);
}

/*********************************************************************/
//display the subtree with given root
/*********************************************************************/
void ConcCEDag::DisplayWithRoot(const int node,list<const ContLoc*> &locStack,list<Action> &actStack,const ProcManager &procManager) const
{
  const Action &act = actLabel[node];
  if(procManager.absLts.GetActs().count(act) == 0) {
    if(trans[node].empty()) {
      DisplayPath(locStack,actStack,procManager);
    } else {
      for(set<int>::const_iterator i = trans[node].begin();i != trans[node].end();i++) {
	DisplayWithRoot(*i,locStack,actStack,procManager);
      }
    }
  } else {
    actStack.push_back(act);
    if(trans[node].empty()) {
      DisplayPath(locStack,actStack,procManager);
    } else {
      for(set<int>::const_iterator i = trans[node].begin();i != trans[node].end();i++) {
	locStack.push_back(locLabel[*i]);
	DisplayWithRoot(*i,locStack,actStack,procManager);
	locStack.pop_back();
      }
    }
    actStack.pop_back();
  }
}

/*********************************************************************/
//display the current path
/*********************************************************************/
void ConcCEDag::DisplayPath(const list<const ContLoc*> &locStack,const list<Action> &actStack,const ProcManager &procManager) const
{
  list<const ContLoc*>::const_iterator i = locStack.begin();
  list<Action>::const_iterator j = actStack.begin();
  bool stutter = false;
  Util::Message(2,"*******************************************\n");
  while((i != locStack.end()) || (j != actStack.end())) {
    stutter = (j != actStack.end()) && (procManager.absLts.GetActs().count(*j) == 0);
    if(i != locStack.end()) {
      if((*i)->GetLocType() == ContLoc::BRANCH) {
	list<const ContLoc*>::const_iterator k = i; k++;
	if(k == locStack.end()) {
	  Util::Message(2,(*i)->ToString() + "\n");
	} else if((*k) == (*i)->GetTSucc()) {
	  Util::Message(2,(*i)->ToString() + " : TRUE\n");
	} else if((*k) == (*i)->GetESucc()) {
	  Util::Message(2,(*i)->ToString() + " : FALSE\n");
	} else {
	  assert(stutter && (*i) == (*k));
	  Util::Message(2,(*i)->ToString() + " : STUTTER\n");
	}
      } else {
	if(stutter) {
	  list<const ContLoc*>::const_iterator k = i; k++;
	  assert((k == locStack.end()) || ((*i) == (*k)));
	  Util::Message(2,(*i)->ToString() + " : STUTTER\n");
	} else {
	  Util::Message(2,(*i)->ToString() + "\n");
	}
      }
      i++;
    }
    if(j != actStack.end()) {
      Util::Message(2,"#### " + j->ToString() + " ####\n");
      stutter = (procManager.absLts.GetActs().count(*j) == 0);
      j++;
    }
  }
  Util::Message(2,"*******************************************\n");
}

/*********************************************************************/
//end of ConcCEDag.cpp
/*********************************************************************/
