/**************************** CPPHeaderFile ***************************

* FileName [Database.h]

* PackageName [main]

* Synopsis [Header file containing all static constants, arguments and
* options to the tool as well as data structures.]

* Description []

* SeeAlso [arc.cpp]

* Author [Sagar Chaki]

* Copyright [ Copyright (c) 2002 by Carnegie Mellon University.  All
* Rights Reserved. This software is for educational purposes only.
* Permission is given to academic institutions to use, copy, and
* modify this software and its documentation provided that this
* introductory message is not removed, that this software and its
* documentation is used for the institutions' internal research and
* educational purposes, and that no monies are exchanged. No guarantee
* is expressed or implied by the distribution of this code. Send
* bug-reports and/or questions to: chaki+@cs.cmu.edu. ]

**********************************************************************/

#ifndef __DATABASE_H__
#define __DATABASE_H__

//other classes needed
class StdcFile;
class Expr;
class FspBody;
class FspTransSeq;
class Action;
class ProcAddInfo;
class ProgInfo;
class LtsTrans;
class FspInfo;
class LtlFormula;
class ProcInfo;
class LtsInfo;
class CallGraph;
class ContLoc;
class ProcManager;
class PredSet;

//needed to interface with the parser
extern FILE *Stdcin;
int Stdcparse();

class Database
{
 private:
  static void DisplayUsage();

 public:

  /*******************************************************************/
  //constants
  /*******************************************************************/
  //the current version of magic
  static const string MAGIC_VERSION;

  //parser related constants for the pre and post increment and
  //decrement operators
  static const short int PRE_INC;
  static const short int PRE_DEC;
  static const short int POST_INC;
  static const short int POST_DEC;

  //the epsilon action
  static const string EPSILON;
  
  //the epsilon specification action. just does an epsilon action
  //and stops.
  static const string EPSILON_SPEC;

  //the name of an epsilon procedure abstraction
  static const string EPSILON_PROC_ABS;
  
  //the default specification LTS abstracting call-sites. this LTS
  //cannot appear in any user-supplied specification. it has just a
  //single transition with the "default-return" action. hence any
  //call-site which is abstracted by this LTS essentially becomes
  //non-simulatable.
  static const string DEFAULT_SPEC;
  
  //the STOP lts state
  static const string STOP_STATE;

  //the ERROR lts state
  static const string ERROR_STATE;

  //a dummy C procedure used as a placeholder when nothing is known
  //about an indirect function call site
  static const string DUMMY_C_PROC;
  
  //various types of conformance checks. currently we support weak
  //simulation, error state reachability, trace containment and LTL
  //model checking
  static const int CONF_SIMUL;
  static const int CONF_REACH;
  static const int CONF_TRACE;
  static const int CONF_LTL;

  //mechanisms for carrying out the conformance check
  static const int CONF_CHECK_EXPLICIT;
  static const int CONF_CHECK_SAT;

  //various C parsers that can be used
  static const int C_PARSER_DEFAULT;
  static const int C_PARSER_CTOOL;
  
  //various theorem provers that can be used
  static const int THEOREM_PROVER_SIMPLIFY;
  static const int THEOREM_PROVER_CPROVER;
  static const int THEOREM_PROVER_CVC;
  static const int THEOREM_PROVER_ICS;
  static const int THEOREM_PROVER_SVC;

  //various SAT solvers that can be used
  static const int SAT_SOLVER_CHAFF;
  static const int SAT_SOLVER_SATO;
  static const int SAT_SOLVER_GRASP;
  static const int SAT_SOLVER_HORN_AI;

  //types of checks for the infeasibility of a path
  static const int PATH_CHECK_FORWARD;
  static const int PATH_CHECK_BACKWARD;

  //various pattern matchers than can be used
  static const int PATTERN_MATCHER_NAIVE;

  //various types of CEGAR strategies - currently we support adding a
  //certain number of predicates that can eliminate the new
  //counterexamples every time, adding a certain number of useful
  //predicates every time and predicate optimization
  static const int CEGAR_ELIMINATE_PRED;
  static const int CEGAR_USEFUL_PRED;
  static const int CEGAR_OPTIMIZE_PRED;
  static const int CEGAR_GREEDY_PRED;

  //the maximum number of bits in a predicate valuation
  static const size_t PVBITS = 24;

  //name of the temporary dot file
  static const string DOT_FILE_NAME;
  
  //parameters for predicate evaluation
  //static const int COVER_NUMBER_TRACES;
  //static const int NUMBER_TRACES;
  //static const int TRACES_LENGTH;
  //static const Random rnd;

  //types of output model formats
  static const int OUTPUT_MODEL_FSP;

  //types of abstraction refinement
  static const int REFINE_NONE;
  static const int REFINE_PRED;
  static const int REFINE_LTS;
  
  /*******************************************************************/
  //arguments and options
  /*******************************************************************/
  //the list of files to be parsed
  static list<string> FILE_NAMES;

  //the name of the abstraction to be checked
  static string ABSTRACTION_NAME;
  
  //the output file for storing the computed simulation relation
  static string SIM_REL_FILE;
  
  //if we just need to parse the input files
  static bool PARSE_ONLY;

  //if the files should be printed while parsing
  static bool PARSE_ECHO;

  //if statistics should be displayed at the end
  static bool DISPLAY_STATISTICS;
  
  //the selected type of conformance check
  static int CONF_TYPE;

  //the selected mechanism for doing the conformance check
  static int CONF_CHECK_MECHANISM;

  //the selected C parser
  static int C_PARSER_TYPE;
  
  //type of theorem prover to be used
  static int THEOREM_PROVER_USED;

  //type of SAT solver to be used
  static int SAT_SOLVER_USED;

  //type of pattern matcher used
  static int PATTERN_MATCHER_USED;

  //whether syntactic checks should be used before calling the theorem
  //prover
  static bool USE_SYNTACTIC_CHECKS;
  
  //whether results of theorem prover calls should be cached
  static bool THEOREM_PROVER_CACHE;

  //the maximum theorem prover cache size
  static int THEOREM_PROVER_CACHE_SIZE;

  //whether to use PAM for alias analysis
  static bool USE_PAM;
  
  //the flag which indicates whether the predicates supplied by the
  //user are to be used or not
  static bool PREDS_FROM_SPEC;
  
  //the verbosity level of the tool
  static int VERBOSITY_LEVEL;

  //whether to use only reachable implementation states
  static bool USE_REACH_IMPL;
  
  //whether the transition relation of the implementation LTS should
  //be precomputed and stored
  static bool PRECOMPUTE_IMPL_TRANS;

  //if the implementation state transitions should be cached
  static bool CACHE_IMPL_TRANS;

  //whether a simulation relation should be created even when using
  //SAT
  static bool SIM_REL_FROM_SAT;
  
  //if pre-optimization phase on just CGA should run
  static bool PREOPTIMIZE;

  //if pre-optimization phase should limit visits to states.
  static bool PREOPTVISIT;

  //limit for number of visits to a state in pre-optimization
  static int PREOPTVISITS;

  //minimum depth to which to perform pre-optimization searches
  static int PREOPTMINDEPTH;

  //minimum number of counterexamples
  static size_t PREOPTMINCES;

  //maximum depth to which to perform pre-optimization searches
  static int PREOPTMAXDEPTH;

  //maximum number of counterexamples
  static size_t PREOPTMAXCES;

  //if traces should be generated based on coverage
  static bool COVERAGE;
  
  //if traces should be generated based on control location vs. impl
  static bool COVERCONT;
  
  //if coverage should be shown (works only for control locations)
  static bool SHOWCOVER;

  //whether to do counter-example guided abstraction refinement
  static bool CEGAR;

  //the type of CEGAR we do
  static int CEGAR_TYPE;

  //whether to do incremental verification
  static bool INC_VERIFY;

  //the maximum number of predicates added in every iteration. this
  //applied only to adding eliminate or useful predicate.
  static int MAX_ADD_PRED;

  //whether we should check the validity of branching in CEs
  static bool CHECK_CE_BRANCHING;

  //what kind of check should be done to check for the feasibility of
  //a path. currently two types are supported - forward and backward.
  static int PATH_CHECK_TYPE;

  //whether to use a pattern matcher during cegar
  static bool USE_PATTERN_MATCHER;

  //the maximum number of global CE dags to be constructed
  static size_t MAX_GLOBAL_CE;

  //the maximum number of backward loops allowed on any path in the
  //global CE dag
  static size_t MAX_CE_LOOP;

  //the probability with which a new node node is created even if
  //there already exists another node with same state labeling
  static double CE_NEW_NODE_PROB;

  //if clauses should be built so that no new variables are introduced
  static bool CNF_NO_NEW_VARS;

  //how large to generate subsets
  static size_t LARGEST_SUBSET;

  //the smallest size of subsets that will be at least tried
  static size_t SMALLEST_SUBSET;

  //cutoff for generating subsets
  static size_t SUBSET_CUTOFF;

  //cutoff for generating conjunctive clauses per spurios CE
  static size_t CONSTRAINT_CUTOFF;

  //if sets of predicates of a larger size should be generated, even if smaller
  //size was sufficient to eliminate.
  static bool STOP_OPT_AFTER_ELIM;

  //if predicate abstraction should be used
  static bool USE_PRED_ABSTRACTION;

  //if LTS abstraction should be used
  static bool USE_LTS_ABSTRACTION;

  //if useless control locations should be removed
  static bool NO_USELESS_LOCS;

  //the maximum number of times a location is allowed to repeat in a
  //predicate inference chain
  static int MAX_PRED_INFER_LOOP;

  //maximum amount of seconds tool should run
  static int MAX_TIME_LIMIT;

  //whether to translate input ltl formulae to buchi automata
  static bool LTL_TO_BUCHI;

  //whether to create any output dot file
  static bool CREATE_DOT_FILE;

  //whether to draw any spec LTSs
  static bool DRAW_LTS;

  //whether to draw the control flow graph
  static bool DRAW_CFG;

  //whether to draw the reduced CFG with inferred predicates
  static bool DRAW_INFERRED_PREDS;

  //whether to draw the LTS obtained after predicate abstraction
  static bool DRAW_PRED_ABS_LTS;

  //whether to draw the LTS obtained after LTS abstraction
  static bool DRAW_LTS_ABS_LTS;

  //whether to draw the LTS obtained after parallel composition
  static bool DRAW_COMPOSED_LTS;

  //draw the CE-DAG obtained
  static bool DRAW_CE_DAG;

  //draw projections of the CE-DAG on LTS abstraction components
  static bool DRAW_CE_PROJ_LTS;

  //draw projections of the CE-DAG on predicate abstraction components
  static bool DRAW_CE_PROJ_PRED;
  
  //whether to output models
  static bool OUTPUT_MODELS;

  //what kind of models are to be output
  static int OUTPUT_MODEL_TYPE;

  /*******************************************************************/
  //data structures
  /*******************************************************************/
  //the command line arguments passed to the tool
  static list<string> commandLine;
  //the canonical names of C files mapped to their asts
  static map<string,StdcFile*> cAsts;
  //the set of spec files already parsed
  static set<string> specFiles;
  //the map of procedure names to additional info specified about them
  static map<string,ProcAddInfo> procAddInfos;
  //the map of program names to info specified about them
  static map<string,ProgInfo> progInfos;
  //the fsp info specified in spec files
  static FspInfo fspInfo;
  //the map of enumeration constants to their values
  static map<string,const BasicExpr*> enums;
  //the names of C procedures mapped to info about them
  static map<string,ProcInfo> procInfos;  
  //the specification LTS name
  static string specLtsName;
  //the name of program to be verified
  static string progName;
  //the root implementation procedure names
  static vector<string> rootProcNames;
  //the guards on the abstraction
  static vector<Expr> absGuards;
  //the vector of LTSs infos
  static vector<LtsInfo> ltsInfos;
  //the mapping from lts infos to their negations
  static map<string,string> ltsNegMap;
  //the call graph
  static CallGraph callGraph;
  //the managers of the root procedures
  static vector<ProcManager> procManagers;
  //the type of abstraction refinement last performed
  static int lastRefType;
  //id of the component whose abstraction was last refined
  static int lastRefComp;
  //the current time
  static double currTime;
  //the priorities of the expressions
  static map<Expr,int> exprPriority;
  //the name of the output file containing the drawn stuff
  static string drawFileName;
  //FILE pointer for the dot output file
  static FILE *dotFilePtr;
  //name of the file containing output FSP models
  static string modelFileName;
  //FILE pointer for the model output file
  static FILE *modelFilePtr;

  /*******************************************************************/
  //methods
  /*******************************************************************/
  static void ParseCommandLine();
  static void ParseFiles();
  static bool IsFileParsed(const string &name);
  static void AddFileAst(const string &name,StdcFile *ast);
  static void ExtractProcInfos();
  static const LtsInfo &GetLtsInfo(const string &name);
  static const LtsInfo &GetLtsNeg(const string &name);
  static LtlFormula GetLtlFormula(const string &name);
  static void ConstructCallGraph();
  static double Checkpoint();
  static const set<Action> &GetSpecActions();
  static set<Action> GetSpecAssignActions(const Expr &lhs,const Expr &rhs,
					  const list<string> &params,
					  const set<Expr> &context);
  static set<Action> GetSpecReturnActions(const Expr &expr,const list<string> &params,
					  const set<Expr> &context);
  static set<Expr> GetSpecRetActExprs();
  static int GetSpecStateNum();
};

#endif //__DATABASE_H__

/*********************************************************************/
//end of Database.h
/*********************************************************************/
