/******************************** CPPFile *****************************

* FileName [LtlChecker.cpp]

* PackageName [main]

* Synopsis [Method definitions of LtlChecker class.]

* SeeAlso [LtlChecker.h]

* Author [Sagar Chaki]

* Copyright [ Copyright (c) 2002 by Carnegie Mellon University. All
* Rights Reserved. This software is for educational purposes only.
* Permission is given to academic institutions to use, copy, and
* modify this software and its documentation provided that this
* introductory message is not removed, that this software and its
* documentation is used for the institutions' internal research and
* educational purposes, and that no monies are exchanged. No guarantee
* is expressed or implied by the distribution of this code. Send
* bug-reports and/or questions to: chaki+@cs.cmu.edu. ]

**********************************************************************/

#include <cassert>
#include <string>
#include <list>
#include <set>
#include <map>
#include <vector>
using namespace std;

#include "Node.h"
#include "Action.h"
#include "GlobalAbsLts.h"
#include "Buchi.h"
#include "LtlChecker.h"

/*********************************************************************/
//define static members
/*********************************************************************/
size_t LtlChecker::ssn;
set<int> LtlChecker::visited1;
set<int> LtlChecker::visited2;
set<int> LtlChecker::stack;
list<int> LtlChecker::statePath;
list<Action> LtlChecker::actPath;
bool LtlChecker::violation;

/*********************************************************************/
//check if the spec LTL formula holds on the program. return true if
//it holds and false otherwise.
/*********************************************************************/
bool LtlChecker::CheckFormula()
{
  set<int> implInit; GlobalAbsLts::GetInitStates(implInit);
  set<int> specInit; Buchi::GetInitStates(specInit);

  ssn = Buchi::GetStateNum();
  visited1.clear();
  visited2.clear();
  stack.clear();
  statePath.clear();
  actPath.clear();
  violation = false;

  for(set<int>::const_iterator i = implInit.begin();(!violation) && (i != implInit.end());i++) {
    for(set<int>::const_iterator j = specInit.begin();(!violation) && (j != specInit.end());j++) {
      if(PropCompatible(*i,*j)) {
	int node = (*i) * ssn + (*j);
	if(visited1.count(node) == 0) DFS1(node);
      }
    }
  }

  visited1.clear();
  visited2.clear();
  stack.clear();
  return !violation;
}

/*********************************************************************/
//the first dfs of ltl model checking
/*********************************************************************/
void LtlChecker::DFS1(const int node)
{
  visited1.insert(node);
  stack.insert(node);
  statePath.push_back(node);
  set<Action> acts; GlobalAbsLts::GetActions(acts);
  for(set<Action>::const_iterator i = acts.begin();(!violation) && (i != acts.end());i++) {
    if(ActCompatible(*i,node % ssn)) {
      set<int> implSuccs; GlobalAbsLts::GetSuccsOnAction(node / ssn,*i,implSuccs);
      set<int> specSuccs = Buchi::GetSuccs(node % ssn);
      for(set<int>::const_iterator j = implSuccs.begin();(!violation) && (j != implSuccs.end());j++) {
	for(set<int>::const_iterator k = specSuccs.begin();(!violation) && (k != specSuccs.end());k++) {
	  if(PropCompatible(*j,*k)) {
	    int next = (*j) * ssn + (*k);
	    if(visited1.count(next) == 0) {
	      actPath.push_back(*i);
	      DFS1(next);
	      if(!violation) actPath.pop_back();
	    }
	  }
	}
      }
    }
  }
  if((!violation) && (Buchi::IsAccept(node % ssn)) && (visited2.count(node) == 0)) {
    DFS2(node);
  }
  if(!violation) {
    stack.erase(node);
    statePath.pop_back();
  }
}

/*********************************************************************/
//the second dfs of ltl model checking
/*********************************************************************/
void LtlChecker::DFS2(const int node)
{
  visited2.insert(node);
  set<Action> acts; GlobalAbsLts::GetActions(acts);
  for(set<Action>::const_iterator i = acts.begin();(!violation) && (i != acts.end());i++) {
    if(ActCompatible(*i,node % ssn)) {
      set<int> implSuccs; GlobalAbsLts::GetSuccsOnAction(node / ssn,*i,implSuccs);
      set<int> specSuccs = Buchi::GetSuccs(node % ssn);
      for(set<int>::const_iterator j = implSuccs.begin();(!violation) && (j != implSuccs.end());j++) {
	for(set<int>::const_iterator k = specSuccs.begin();(!violation) && (k != specSuccs.end());k++) {
	  if(PropCompatible(*j,*k)) {
	    int next = (*j) * ssn + (*k);
	    if(stack.count(next) != 0) {
	      statePath.push_back(next);
	      actPath.push_back(*i);
	      violation = true;
	      return;
	    }
	    if((!violation) && (visited2.count(next) == 0)) {
	      statePath.push_back(next);
	      actPath.push_back(*i);
	      DFS2(next);
	      if(!violation) {
		statePath.pop_back();
		actPath.pop_back();
	      }
	    }
	  }
	}
      }
    }
  }
}

/*********************************************************************/
//return true if an implementation state is propositionally compatible
//with a specification state. the two states are arguments.
/*********************************************************************/
bool LtlChecker::PropCompatible(int implState,int specState)
{
  const set<Expr> implProps = GlobalAbsLts::GetPropositions(implState);
  const set<Expr> posProps = Buchi::GetPosProps(specState);
  const set<Expr> negProps = Buchi::GetNegProps(specState);

  for(set<Expr>::const_iterator i = posProps.begin();i != posProps.end();i++) {
    if(implProps.count(*i) == 0) return false;
  }
  for(set<Expr>::const_iterator i = negProps.begin();i != negProps.end();i++) {
    if(implProps.count(*i) != 0) return false;
  }  
  return true;
}

/*********************************************************************/
//return true if the specification state is compatible with the action
//i.e. either the action belongs to the label or its negation does not
/*********************************************************************/
bool LtlChecker::ActCompatible(const Action &act,int specState)
{
  const set<Action> posActs = Buchi::GetPosActs(specState);
  const set<Action> negActs = Buchi::GetNegActs(specState);
  if(posActs.empty()) {
    for(set<Action>::const_iterator i = negActs.begin();i != negActs.end();i++) {
      if(act == (*i)) return false;
    }
    return true;
  } else {
    assert(posActs.size() == 1);
    return (act == *(posActs.begin()));
  }
}

/*********************************************************************/
//end of LtlChecker.cpp
/*********************************************************************/
