/******************************** CPPFile *****************************

* FileName [LtsInfo.cpp]

* PackageName [main]

* Synopsis [Method definitions of LtsInfo class.]

* SeeAlso [LtsInfo.h]

* Author [Sagar Chaki]

* Copyright [ Copyright (c) 2002 by Carnegie Mellon University. All
* Rights Reserved. This software is for educational purposes only.
* Permission is given to academic institutions to use, copy, and
* modify this software and its documentation provided that this
* introductory message is not removed, that this software and its
* documentation is used for the institutions' internal research and
* educational purposes, and that no monies are exchanged. No guarantee
* is expressed or implied by the distribution of this code. Send
* bug-reports and/or questions to: chaki+@cs.cmu.edu. ]

**********************************************************************/

#include <cstdio>
#include <cassert>
#include <string>
#include <list>
#include <set>
#include <map>
#include <vector>
using namespace std;

#include "Node.h"
#include "Action.h"
#include "ActionManager.h"
#include "LtsTrans.h"
#include "LtsInfo.h"
#include "Database.h"

/*********************************************************************/
//constructors and destructors
/*********************************************************************/
LtsInfo::LtsInfo()
{
  useful = false;
  index = 0;
}

LtsInfo::LtsInfo(const string &n,const set<string> &s,const set<Action> &a,const set<LtsTrans> &t)
{
  name = n;
  useful = false;
  states = s;
  actions = a;
  trans = t;
  index = 0;

  //create the states to indices map
  int id = 0;
  for(set<string>::const_iterator i = states.begin();i != states.end();i++,id++) {
    statesToIds[*i] = id;
  }

  //create the successor and predecessor action map
  for(set<LtsTrans>::const_iterator i = trans.begin();i != trans.end();i++) {
    succMap[pair<string,Action>(i->GetInitState(),i->GetAction())].insert(i->GetFinalState());
    predMap[pair<string,Action>(i->GetFinalState(),i->GetAction())].insert(i->GetInitState());
  }
}

LtsInfo::LtsInfo(const LtsInfo &rhs) 
{ 
  *this = rhs; 
}

/*********************************************************************/
//operators
/*********************************************************************/
const LtsInfo &LtsInfo::operator = (const LtsInfo &rhs)
{
  name = rhs.name;
  useful = rhs.useful;
  states = rhs.states;
  statesToIds = rhs.statesToIds;
  actions = rhs.actions;
  trans = rhs.trans;
  index = rhs.index;
  succMap = rhs.succMap;
  predMap = rhs.predMap;
  return *this;
}

/*********************************************************************/
//compute the set of useful states
/*********************************************************************/
void LtsInfo::ComputeUseful()
{
  //if this is the info about the spec LTS then it is useful
  if(name == Database::specLtsName) {
    useful = true;
    return;
  }

  //check if this lts has any return action other than the default
  //action or assign actions
  for(set<Action>::iterator i = actions.begin();i != actions.end();i++) {
    if((!i->IsDefaultRet() && (i->IsReturn())) || 
       (i->IsAssign())) {
      useful = true;
      return;
    }
  }
  
  //check if it has some action common to the spec LTS
  const set<Action> &specActs = Database::GetSpecActions();
  for(set<Action>::iterator i = actions.begin();i != actions.end();i++) {
    //only consider basic non-epsilon actions
    if((!i->IsBasic()) || (i->IsSpecEpsilon()))continue;
    for(set<Action>::const_iterator j = specActs.begin();j != specActs.end();j++) {
      if(*i == *j) {
	useful = true;
	return;
      }
    }
  }
}

/*********************************************************************/
//return the expressions associated with return actions
/*********************************************************************/
set<Expr> LtsInfo::GetRetActExprs() const
{
  set<Expr> res;
  for(set<Action>::iterator i = actions.begin();i != actions.end();i++) {
    if((i->IsReturn()) && (!i->IsWildcardRet())) {
      res.insert(i->GetRetExpr());
    }
  }
  return res;
}

/*********************************************************************/
//return set of states
/*********************************************************************/
const set<string> &LtsInfo::GetStates() const
{
  return states;
}

/*********************************************************************/
//return set of actions
/*********************************************************************/
const set<Action> &LtsInfo::GetActions() const
{
  return actions;
}

/*********************************************************************/
//compute the set of actions associated with a return. the arguments
//are the expression for the return value, a list of arguments to the
//procedure, and a list of expressions describing the state in which
//the return is executed.
/*********************************************************************/
set<Action> LtsInfo::GetReturnActions(const Expr &expr,const list<string> &params,
				      const set<Expr> &context) const
{
  set<Action> res;
  set<Action> returns = ComputeReturnActions();

  //check for wildcard returns
  bool hasWild = false;
  if(returns.count(ActionManager::GetWildcardRetAction()) != 0) {
    hasWild = true;
    returns.erase(ActionManager::GetWildcardRetAction());
  }

  for(set<Action>::const_iterator i = returns.begin();i != returns.end();i++) {
    if(i->ReturnActionPossible(expr,params,context)) {
      res.insert(*i);
    }
  }

  //if there is a wildcard and nothing has matched yet check if a
  //wildcard should match
  if(hasWild && res.empty()) {
    if(ActionManager::GetWildcardRetAction().ReturnActionPossible(expr,params,context)) {
      res.insert(ActionManager::GetWildcardRetAction());
    }
  }

  return res;
}

/*********************************************************************/
//compute the set of actions associated with a return. the arguments
//are the expression for the return value, a list of arguments to the
//procedure, and a list of expressions describing the state in which
//the return is executed.
/*********************************************************************/
set<Action> LtsInfo::GetAssignActions(const Expr &lhs,const Expr &rhs,
				      const list<string> &params,
				      const set<Expr> &context) const
{
  set<Action> res;
  set<Action> assigns = ComputeAssignActions();
  for(set<Action>::const_iterator i = assigns.begin();i != assigns.end();i++) {
    if(i->AssignActionPossible(lhs,rhs,params,context)) {
      res.insert(*i);
    }
  }
  return res;
}

/*********************************************************************/
//compute the set of return actions in this LTS
/*********************************************************************/
set<Action> LtsInfo::ComputeReturnActions() const
{
  set<Action> res;
  for(set<Action>::const_iterator i = actions.begin();i != actions.end();i++) {
    if(i->IsReturn()) {
      res.insert(*i);
    }
  }
  return res;
}

/*********************************************************************/
//compute the set of assign actions in this LTS
/*********************************************************************/
set<Action> LtsInfo::ComputeAssignActions() const
{
  set<Action> res;
  for(set<Action>::const_iterator i = actions.begin();i != actions.end();i++) {
    if(i->IsAssign()) {
      res.insert(*i);
    }
  }
  return res;
}

/*********************************************************************/
//given a state of the lts return the list of its outgoing transitions
/*********************************************************************/
list<LtsTrans> LtsInfo::GetOutTrans(const string &source) const
{
  list<LtsTrans> res;
  for(set<LtsTrans>::const_iterator i = trans.begin();i != trans.end();i++) {
    if(i->GetInitState() == source) {
      res.push_back(*i);
    }
  }
  return res;
}

/*********************************************************************/
//given a state X and an action Y return the set of states Z such that
//there is transition from X to Z with action Y.
/*********************************************************************/
set<string> LtsInfo::GetSuccsOnAction(const string &source,const Action &action) const
{
  //if the action is not a default return and does not belong to
  //the set of actions of this LTS then turn it into an epsilon
  //action
  if((!action.IsDefaultRet()) && (actions.count(action) == 0)) {
    map< pair<string,Action>,set<string> >::const_iterator i = 
      succMap.find(pair<string,Action>(source,ActionManager::GetSpecEpsilonAction()));
    if(i == succMap.end()) return set<string>();
    else return i->second;
  } else {
    map< pair<string,Action>,set<string> >::const_iterator i = 
      succMap.find(pair<string,Action>(source,action));
    if(i == succMap.end()) return set<string>();
    else return i->second;
  }
}

/*********************************************************************/
//given a state X and an action Y return the set of states Z such that
//there is transition from X to Z with action Y.
/*********************************************************************/
set<string> LtsInfo::GetSuccsOnAction(const int source,const Action &action) const
{
  return GetSuccsOnAction(GetIdState(source),action);
}

/*********************************************************************/
//given a state X and an action Y return the set of states Z such that
//there is transition from Z to X with action Y.
/*********************************************************************/
set<string> LtsInfo::GetPredsOnAction(const string &source,const Action &action) const
{
  //if the action is not a default return and does not belong to
  //the set of actions of this LTS then turn it into an epsilon
  //action
  if((!action.IsDefaultRet()) && (actions.count(action) == 0)) {
    map< pair<string,Action>,set<string> >::const_iterator i = 
      predMap.find(pair<string,Action>(source,ActionManager::GetSpecEpsilonAction()));
    if(i == predMap.end()) return set<string>();
    else return i->second;
  } else {
    map< pair<string,Action>,set<string> >::const_iterator i = 
      predMap.find(pair<string,Action>(source,action));
    if(i == predMap.end()) return set<string>();
    else return i->second;
  }
}

/*********************************************************************/
//given a state X and an action Y return the set of states Z such that
//there is transition from Z to X with action Y.
/*********************************************************************/
set<string> LtsInfo::GetPredsOnAction(const int source,const Action &action) const
{
  return GetPredsOnAction(GetIdState(source),action);
}

/*********************************************************************/
//return the id of a state
/*********************************************************************/
int LtsInfo::GetStateId(const string &s) const
{
  map<string,int>::const_iterator i = statesToIds.find(s);
  assert(i != statesToIds.end());
  return i->second;
}

/*********************************************************************/
//given an id return the corresponding state
/*********************************************************************/
string LtsInfo::GetIdState(const int id) const
{
  for(map<string,int>::const_iterator i = statesToIds.begin();i != statesToIds.end();i++) {
    if(i->second == id) return i->first;
  }
  assert(false);
  return "";
}

/*********************************************************************/
//draw the LTS as a dot file
/*********************************************************************/
void LtsInfo::Draw() const
{
  FILE *out = Database::dotFilePtr;
  fprintf(out,"  //FSP process %s\n",name.c_str());
  fprintf(out,"  subgraph cluster_FSP_%s {\n",name.c_str());  
  fprintf(out,"    color = white;\n");
  fprintf(out,"    %s_%s [peripheries = 2];\n",name.c_str(),name.c_str()); 
  for(set<string>::const_iterator i = states.begin();i != states.end();i++) {
    fprintf(out,"    %s_%s [label = \"%s\"];\n",name.c_str(),i->c_str(),i->c_str()); 
  }
  for(set<LtsTrans>::const_iterator i = trans.begin();i != trans.end();i++) {
    fprintf(out,"    %s_%s -> %s_%s [label = \"%s\"];\n",name.c_str(),i->GetInitState().c_str(),name.c_str(),i->GetFinalState().c_str(),i->GetAction().ToString().c_str());
  }
  fprintf(out,"    label = \"FSP process: %s\";\n",name.c_str());
  fprintf(out,"    labelloc = \"b\";\n");
  fprintf(out,"  }\n\n");
}

/*********************************************************************/
//end of LtsInfo.cpp
/*********************************************************************/
