/******************************** CPPFile *****************************

* FileName [Statistics.cpp]

* PackageName [main]

* Synopsis [Method definitions of Statistics class.]

* SeeAlso [Statistics.h]

* Author [Sagar Chaki]

* Copyright [ Copyright (c) 2002 by Carnegie Mellon University. All
* Rights Reserved. This software is for educational purposes only.
* Permission is given to academic institutions to use, copy, and
* modify this software and its documentation provided that this
* introductory message is not removed, that this software and its
* documentation is used for the institutions' internal research and
* educational purposes, and that no monies are exchanged. No guarantee
* is expressed or implied by the distribution of this code. Send
* bug-reports and/or questions to: chaki+@cs.cmu.edu. ]

**********************************************************************/

#include <cstdio>
#include <cassert>
#include <sys/time.h>
#ifndef WIN32
#include <sys/resource.h>
#endif //WIN32
#include <unistd.h>
#include <string>
#include <list>
#include <set>
#include <vector>
#include <map>
#include <bitset>
using namespace std;

#include "Util.h"
#include "Statistics.h"
#include "Node.h"
#include "Action.h"
#include "Database.h"
#include "Predicate.h"
#include "PredSet.h"
#include "ContLoc.h"
#include "LocalAbsLts.h"
#include "ConcCEDag.h"
#include "ProcManager.h"
#include "ContLoc.h"

/*********************************************************************/
//static members
/*********************************************************************/
double Statistics::startTime = 0.0;
double Statistics::endTime = 0.0;
size_t Statistics::elimCombNum = 0;
size_t Statistics::maxElimCombNum = 0;
size_t Statistics::maxElimCombSize = 0;
size_t Statistics::maxTriedCombSize = 0;
int Statistics::seedNum = 0;
int Statistics::stateNum = 0;
int Statistics::iterNum = 0;
int Statistics::predIterNum = 0;
int Statistics::ltsIterNum = 0;
int Statistics::tpCalls = 0;
int Statistics::tpInCache = 0;
int Statistics::tpCacheSize = 0;

/*********************************************************************/
//display statistics
/*********************************************************************/
void Statistics::Display()
{
  //checkpoint to get the correct time
  Database::Checkpoint();

  if(Database::DISPLAY_STATISTICS) {
#ifdef WIN32
    double cpuTime = static_cast<double>(clock()) / CLOCKS_PER_SEC;
#else
    struct rusage usage;
    getrusage(RUSAGE_SELF,&usage);
    double cpuTime = usage.ru_utime.tv_sec + usage.ru_utime.tv_usec / 1000000.0 +
      usage.ru_stime.tv_sec + usage.ru_stime.tv_usec / 1000000.0;
#endif //WIN32
    fprintf(stderr,"total global time = %.1f milliseconds\n",(endTime - startTime) * 1000);
    fprintf(stderr,"total cpu time = %.1f milliseconds\n",cpuTime * 1000);
    fprintf(stderr,"total number of eliminating combinations = %d\n",elimCombNum);
    fprintf(stderr,"max number of eliminating combinations for a CE = %d\n",maxElimCombNum);
    fprintf(stderr,"max size of eliminating combination = %d\n",maxElimCombSize);
    fprintf(stderr,"max size of tried combination = %d\n",maxTriedCombSize);
    fprintf(stderr,"number of iterations = %d\n",iterNum);
    fprintf(stderr,"number of predicate iterations = %d\n",predIterNum);
    fprintf(stderr,"number of lts iterations = %d\n",ltsIterNum);
    fprintf(stderr,"number of seed branches : %d\n",seedNum);
    fprintf(stderr,"number of states : %d\n",stateNum);
    fprintf(stderr,"number of theorem prover calls = %d\n",tpCalls);
    fprintf(stderr,"number of formulas found in cache = %d\n",tpInCache);
    fprintf(stderr,"number of formulas stored in cache = %d\n",tpCacheSize);
    
    Util::Message(2,"total global time = %.1f milliseconds\n",(endTime - startTime) * 1000);
    Util::Message(2,"total cpu time = %.1f milliseconds\n",cpuTime * 1000);
    Util::Message(2,"total number of eliminating combinations = %d\n",elimCombNum);
    Util::Message(2,"max number of eliminating combinations for a CE = %d\n",maxElimCombNum);
    Util::Message(2,"max size of eliminating combination = %d\n",maxElimCombSize);
    Util::Message(2,"max size of tried combination = %d\n",maxTriedCombSize);
    Util::Message(2,"number of iterations = %d\n",iterNum);
    Util::Message(2,"number of predicate iterations = %d\n",predIterNum);
    Util::Message(2,"number of lts iterations = %d\n",ltsIterNum);
    Util::Message(2,"number of seed branches : %d\n",seedNum);
    Util::Message(2,"number of states : %d\n",stateNum);
    Util::Message(2,"number of theorem prover calls = %d\n",tpCalls); 
    Util::Message(2,"number of formulas found in cache = %d\n",tpInCache);
    Util::Message(2,"number of formulas stored in cache = %d\n",tpCacheSize);
    
    //print expressions and predicates used
    for(size_t k = 0;k < Database::procManagers.size();k++) {
      const vector < ContLoc*> &allLocs = Database::procManagers[k].GetUsefulLocs();
      set <Expr> es;
      PredSet ap;
      
      for(vector < ContLoc* >::const_iterator i = allLocs.begin ();
	  i != allLocs.end (); i++) {
	const PredSet &p = (*i)->GetPreds ();
	ap.Merge (p);
	const vector <Predicate> &pv = p.GetPreds ();
	for (vector <Predicate>::const_iterator j = pv.begin ();
	     j != pv.end (); j++) {
	  const vector <Expr> &pev = (*j).GetValToExpr ();
	  es.insert(pev.begin(),pev.end());
	}
      }
      
      fprintf(stderr, "Using %i expressions:\n", es.size ());
      for (set <Expr>::const_iterator i = es.begin();
	   i != es.end (); i++) {
	fprintf(stderr, "  ");
	fprintf(stderr, (*i).ToString ().c_str ());
	fprintf(stderr, "\n");
      }
      fprintf(stderr, "Using %i predicates:\n", ap.GetPreds ().size ());
      fprintf(stderr, ap.ToString ().c_str ());
      fprintf(stderr, "\n");
    }
  }
}

/*********************************************************************/
//end of Statistics.cpp
/*********************************************************************/
