/******************************** CPPFile *****************************

* FileName [magic.cpp]

* PackageName [main]

* Synopsis [The file containing the main procedure.]

* SeeAlso []

* Author [Sagar Chaki]

* Copyright [ Copyright (c) 2002 by Carnegie Mellon University. All
* Rights Reserved. This software is for educational purposes only.
* Permission is given to academic institutions to use, copy, and
* modify this software and its documentation provided that this
* introductory message is not removed, that this software and its
* documentation is used for the institutions' internal research and
* educational purposes, and that no monies are exchanged. No guarantee
* is expressed or implied by the distribution of this code. Send
* bug-reports and/or questions to: chaki+@cs.cmu.edu. ]

**********************************************************************/

#include <cstdio>
#include <cassert>
#include <signal.h>
#include <time.h>
#include <string>
#include <list>
#include <map>
#include <set>
#include <vector>
#include <bitset>
using namespace std;

#include "Util.h"
#include "Statistics.h"
#include "Node.h"
#include "Action.h"
#include "Database.h"
#include "Predicate.h"
#include "PredSet.h"
#include "TheoremProver.h"
#include "SimChecker.h"
#include "InputProcessor.h"
#include "ModelExtractor.h"
#include "ProofGen.h"
#include "CEGen.h"
#include "ConcCEDag.h"
#include "AbsRefiner.h"
#include "PredAbsRefiner.h"
#include "LocalAbsLts.h"
#include "ProcManager.h"
#include "GlobalAbsLts.h"
#ifndef MAGIC_LITE
#include "PamAPI.h"
using namespace ctool;
#endif //MAGIC_LITE
#include "LtlChecker.h"

/*********************************************************************/
//procedure declarations
/*********************************************************************/
void InstallHandlers();
__attribute__ ((noreturn)) void GlobalShutdown(int sig);

/*********************************************************************/
//this is where it all begins
/*********************************************************************/
int main(int argc,char *argv[])
{
  //install the signal handlers
  InstallHandlers();

  //set random number generator seed
  srand(time(NULL));

  //copy the command line to the database and also print the command
  //line arguments
  Util::Message(2,"MAGIC version %s ...\n",Database::MAGIC_VERSION.c_str());
  Util::Message(2,"Command Line: %s ",argv[0]);
  for(int i = 1;i < argc;i++) {
    Util::Message(2,"%s ",argv[i]);
    Database::commandLine.push_back(argv[i]);
  }
  Util::Message(2,"\n");

  //run the input processor
  Database::Checkpoint();
  bool res1 = InputProcessor::Run();
  Util::Message(2,"inputs processed in %.1f milliseconds ...\n",Database::Checkpoint() * 1000);

  //if there is any abstraction to be verified then proceed with the
  //abstract-verify-refine cycle
  bool res3 = false;
  if(res1) {
    if(Database::CONF_TYPE == Database::CONF_SIMUL) {
      Util::Message(2,"checking " + Database::specLtsName + " {simul} " + Database::progName + "\n");
    } else if(Database::CONF_TYPE == Database::CONF_REACH) {
      Util::Message(2,"checking " + Database::specLtsName + " {reach} " + Database::progName + "\n");
    } else if(Database::CONF_TYPE == Database::CONF_LTL) {
      Util::Message(2,"checking " + Database::specLtsName + " {LTL} " + Database::progName + "\n");
    } else assert(false);

    if (Database::PREOPTIMIZE)
      (PredAbsRefiner(Database::procManagers[0])).PreOptimize ();

    while(true) {
      Util::Message(2,"starting iteration number %d ...\n",++(Statistics::iterNum));

      //update the number of predicate or lts iterations
      if((Statistics::predIterNum == 0) && (Statistics::ltsIterNum == 0)) {
	Statistics::predIterNum++;
      } else {
	bool predIter = false;
	for(size_t i = 0;i < Database::procManagers.size();i++) {
	  if(Database::procManagers[i].GetNewSeeds()) {
	    Statistics::predIterNum++;
	    predIter = true;
	    break;
	  }
	}
	if(!predIter) Statistics::ltsIterNum++;
      }

      //update the number of seed branches
      Statistics::seedNum = 0;
      for(size_t i = 0;i < Database::procManagers.size();i++) {
	Statistics::seedNum += Database::procManagers[i].GetSeedBranches().size();
      }

      //extract the model
      bool res2 = ModelExtractor::Run();
      Util::Message(2,"implementation machine extracted in %.1f milliseconds ...\n",
		    Database::Checkpoint() * 1000);
      if(!res2) Util::Error("error while extracting model ...\n");

      //update the maximum number of global states
      Statistics::stateNum = (GlobalAbsLts::GetStateNum() > Statistics::stateNum) ? GlobalAbsLts::GetStateNum() : Statistics::stateNum;

      //check if time limit has been exceeded
      if((Statistics::endTime - Statistics::startTime) > Database::MAX_TIME_LIMIT) {
	Util::Error("time limit exceeded ...\n");
      }
      
      //verify simulation
      if(Database::CONF_TYPE == Database::CONF_SIMUL) {
	res3 = SimChecker::CheckSimulationProperty();
	Util::Message(2,"simulation checked in %.1f milliseconds ...\n",
		      Database::Checkpoint() * 1000);
      }
      //verify error state reachability
      else if(Database::CONF_TYPE == Database::CONF_REACH) {
	res3 = SimChecker::CheckReachability();
	Util::Message(2,"reachability checked in %.1f milliseconds ...\n",
		      Database::Checkpoint() * 1000);
      }
      //verify LTL formula
      else if(Database::CONF_TYPE == Database::CONF_LTL) {
	res3 = LtlChecker::CheckFormula();
	Util::Message(2,"LTL formula checked in %.1f milliseconds ...\n",
		      Database::Checkpoint() * 1000);
      }
      //impossible
      else assert(false);
      
      //if the abstraction holds generate the proof and exit
      if(res3) {
	Util::Message(1,"conformance relation exists !!\n");
	ProofGen::Run();
	break;
      }
      //if the abstraction does not hold, generate counter-example,
      //verify it and refine if necessary.
      else {
	if(!Database::CEGAR) {
	  CEGen::Run();
	  Util::Message(1,"conformance relation does not exist !!\n");
	  break;
	} else {
	  //refine the abstraction on the basis of the
	  //counter-example. if the counter example is valid the
	  //simulation does not hold. otherwise we repeat the loop if
	  //new predicates were discovered.
	  int res4 = AbsRefiner::Run();
	  if(res4 == AbsRefiner::CE_VALID) {
	    Util::Message(1,"conformance relation does not exist !!\n");
	    break;
	  } else {
	    Util::Message(1,"spurious counter-example detected !!\n");
	    if(res4 == AbsRefiner::CE_INVALID_NOT_REFINED) {
	      Util::Message(1,"abstraction not refined !!\n");
	      break;
	    } else {
	      assert(res4 == AbsRefiner::CE_INVALID_REFINED);
	      Util::Message(1,"abstraction refined !!\n");
	    }
	  }
	}
      }
    }
  }
  
  //all done
  if(Database::ABSTRACTION_NAME == "") {
    Util::Message(1,"no abstraction specified ...\n");
  } else if(res1) {
    if(res3) {
      Util::Message(1,"abstraction " + Database::ABSTRACTION_NAME + " is valid ...\n");
    } else {
      Util::Message(1,"abstraction " + Database::ABSTRACTION_NAME + " is invalid ...\n");
    }
  } else {
    Util::Message(1,"abstraction " + Database::ABSTRACTION_NAME + " not found ...\n");
  }

  //cleanup
  GlobalShutdown(-1);

  return 0;
}

/*********************************************************************/
//install signal handlers
/*********************************************************************/
void InstallHandlers()
{
#ifdef WIN32
  signal(SIGABRT,GlobalShutdown);
  signal(SIGFPE,GlobalShutdown);
  signal(SIGILL,GlobalShutdown);
  signal(SIGINT,GlobalShutdown);
  signal(SIGSEGV,GlobalShutdown);
  signal(SIGTERM,GlobalShutdown);
#else
  struct sigaction sa;
  sa.sa_handler = GlobalShutdown;
  sigaction(SIGINT,&sa,NULL);
  sigaction(SIGSEGV,&sa,NULL);
  sigaction(SIGILL,&sa,NULL);
  sigaction(SIGFPE,&sa,NULL);
  sigaction(SIGBUS,&sa,NULL);
  sigaction(SIGTERM,&sa,NULL);
  sigaction(SIGABRT,&sa,NULL);
  sigaction(SIGSYS,&sa,NULL);
#endif //WIN32
}

/*********************************************************************/
//cleanup
/*********************************************************************/
__attribute__ ((noreturn)) void GlobalShutdown(int sig)
{
  //display statistics
  Statistics::Display();

  //shutdown the theorem prover
  TheoremProver::Shutdown();

#ifndef MAGIC_LITE
  //cleanup the PAM interface
  PamAPI::Cleanup();
#endif //MAGIC_LITE

  //close the output dot file if any
  if(Database::dotFilePtr != NULL) {
    fprintf(Database::dotFilePtr,"}\n");
    fclose(Database::dotFilePtr);
    string command = "dot -Tps " + Database::DOT_FILE_NAME + " -o " + Database::drawFileName;
    system(command.c_str());
    remove(Database::DOT_FILE_NAME.c_str());
  }

  //close the output FSP file if any
  if(Database::modelFilePtr != NULL) {
    fclose(Database::modelFilePtr);
  }

  //cleanup the procedure managers
  for(vector<ProcManager>::const_iterator i = Database::procManagers.begin();
      i != Database::procManagers.end();i++) {
    i->Cleanup();
  }

  //display a message
  if(sig == -1) {
    Util::Message(0,"terminating normally ...\n");
  } else {
    Util::Message(0,"received signal %d ...\n",sig);
  }

  //all done
  exit(0);
}

/*********************************************************************/
//end of magic.cpp
/*********************************************************************/
