/******************************** CPPFile *****************************

* FileName [SimplifyTP.cpp]

* PackageName [theorem-prover.simplify]

* Synopsis [This file contains the methods for the SimplifyTP class.]

* SeeAlso [SimplifyTP.h]

* Author [Sagar Chaki]

* Copyright [ Copyright (c) 2002 by Carnegie Mellon University. All
* Rights Reserved. This software is for educational purposes only.
* Permission is given to academic institutions to use, copy, and
* modify this software and its documentation provided that this
* introductory message is not removed, that this software and its
* documentation is used for the institutions' internal research and
* educational purposes, and that no monies are exchanged. No guarantee
* is expressed or implied by the distribution of this code. Send
* bug-reports and/or questions to: chaki+@cs.cmu.edu. ]

**********************************************************************/

#include <cstdio>
#include <cassert>
#include <sys/types.h>
#ifndef WIN32
#include <sys/wait.h>
#endif //WIN32
#include <unistd.h>
#include <signal.h>
#include <string>
#include <set>
#include <list>
#include <map>
#include <vector>
#include <typeinfo>
using namespace std;

#include "Util.h"
#include "Statistics.h"
#include "Node.h"
#include "GenericTP.h"
#include "SimplifyTP.h"
#include "Database.h"
#include "StdcParser.h"

/*********************************************************************/
//static data
/*********************************************************************/
const string SimplifyTP::FIELD_ACCESS = "extract";
const string SimplifyTP::POINTER_DEREF = "deref";
const string SimplifyTP::ADDRESS = "address";
const string SimplifyTP::BITWISE_NOT = "bwnot";
const string SimplifyTP::BITWISE_AND = "bwand";
const string SimplifyTP::BITWISE_XOR = "bwxor";
const string SimplifyTP::BITWISE_OR = "bwor";
const string SimplifyTP::LEFT_SHIFT = "lsh";
const string SimplifyTP::RIGHT_SHIFT = "rsh";
const string SimplifyTP::PROCEDURE = "proc_";
const string SimplifyTP::ALPHA_RENAMED = "alpha_renamed";
const size_t SimplifyTP::MAX_TOTAL_QUERY_SIZE = 100000000;

/*********************************************************************/
//initialise the theorem prover
/*********************************************************************/
void SimplifyTP::Initialise()
{
#ifdef WIN32
  HANDLE hChildStdinRd,hChildStdinWr,hChildStdoutRd,hChildStdoutWr; 
  HANDLE hSaveStdin,hSaveStdout;
  SECURITY_ATTRIBUTES saAttr; 
  BOOL fSuccess; 
  //Set the bInheritHandle flag so pipe handles are inherited. 
  saAttr.nLength = sizeof(SECURITY_ATTRIBUTES); 
  saAttr.bInheritHandle = TRUE; 
  saAttr.lpSecurityDescriptor = NULL; 
  //The steps for redirecting child process's STDOUT: 
  //    1. Save current STDOUT, to be restored later. 
  //    2. Create anonymous pipe to be STDOUT for child process. 
  //    3. Set STDOUT of the parent process to be write handle to 
  //       the pipe, so it is inherited by the child process. 
  //    4. Create a noninheritable duplicate of the read handle and
  //       close the inheritable read handle. 
  //Save the handle to the current STDOUT. 
  hSaveStdout = GetStdHandle(STD_OUTPUT_HANDLE); 
  //Create a pipe for the child process's STDOUT. 
  if (! CreatePipe(&hChildStdoutRd, &hChildStdoutWr, &saAttr, 0)) {
    Util::Error("Stdout pipe creation failed\n"); 
  }
  //Set a write handle to the pipe to be STDOUT.    
  if (! SetStdHandle(STD_OUTPUT_HANDLE, hChildStdoutWr)) {
    Util::Error("Redirecting STDOUT failed");
  }
  //Create noninheritable read handle and close the inheritable read
  //handle.  
  fSuccess = DuplicateHandle(GetCurrentProcess(), hChildStdoutRd,
			     GetCurrentProcess(), &in , 0,
			     FALSE,
			     DUPLICATE_SAME_ACCESS);
  if( !fSuccess ) Util::Error("DuplicateHandle failed");
  CloseHandle(hChildStdoutRd);
  //The steps for redirecting child process's STDIN: 
  //    1.  Save current STDIN, to be restored later. 
  //    2.  Create anonymous pipe to be STDIN for child process. 
  //    3.  Set STDIN of the parent to be the read handle to the 
  //        pipe, so it is inherited by the child process. 
  //    4.  Create a noninheritable duplicate of the write handle, 
  //        and close the inheritable write handle. 
  //Save the handle to the current STDIN.
  hSaveStdin = GetStdHandle(STD_INPUT_HANDLE); 
  //Create a pipe for the child process's STDIN. 
  if (! CreatePipe(&hChildStdinRd, &hChildStdinWr, &saAttr, 0)) {
    Util::Error("Stdin pipe creation failed\n"); 
  }
  //Set a read handle to the pipe to be STDIN. 
  if (! SetStdHandle(STD_INPUT_HANDLE, hChildStdinRd)) {
    Util::Error("Redirecting Stdin failed"); 
  }
  //Duplicate the write handle to the pipe so it is not inherited. 
  fSuccess = DuplicateHandle(GetCurrentProcess(), hChildStdinWr, 
			     GetCurrentProcess(), &out, 0, 
			     FALSE,                  //not inherited 
			     DUPLICATE_SAME_ACCESS); 
  if (! fSuccess) Util::Error("DuplicateHandle failed"); 
  CloseHandle(hChildStdinWr); 
  //Now create the child process.
  STARTUPINFO siStartInfo;
  BOOL bFuncRetn = FALSE; 
  //Set up members of the PROCESS_INFORMATION structure. 
  ZeroMemory( &proc, sizeof(PROCESS_INFORMATION) );
  //Set up members of the STARTUPINFO structure. 
  ZeroMemory( &siStartInfo, sizeof(STARTUPINFO) );
  siStartInfo.cb = sizeof(STARTUPINFO);
  //create the name of simplify executable file
  char *envptr = getenv("MAGICROOT");
  if(envptr == NULL) Util::Error("MAGICROOT not set ... Aborting ...\n");
  string command = string("\"") + string(envptr) + "\\theorem-prover\\simplify\\Simplify.exe\" -nosc";
  char commLine[1024];
  strcpy(commLine,command.c_str());
  //Create the child process. 
  bFuncRetn = CreateProcess(NULL, 
			    commLine,     //command line 
			    NULL,         //process security attributes 
			    NULL,         //primary thread security attributes 
			    TRUE,         //handles are inherited 
			    0,            //creation flags 
			    NULL,         //use parent's environment 
			    NULL,         //use parent's current directory 
			    &siStartInfo, //STARTUPINFO pointer 
			    &proc);       //receives PROCESS_INFORMATION 
  if (bFuncRetn == 0) Util::Error("CreateProcess failed\n");
  else Util::Message(2,"Simplify process created ...\n");
  //restore the saved STDIN and STDOUT. 
  if (! SetStdHandle(STD_INPUT_HANDLE, hSaveStdin)) Util::Error("Re-redirecting Stdin failed\n"); 
  if (! SetStdHandle(STD_OUTPUT_HANDLE, hSaveStdout)) Util::Error("Re-redirecting Stdout failed\n");
  //set the query size and clear enum constraints
  totalQuerySize = 0;
  enumsAdded.clear();
  //read up the initial Simplify prompt
  ReadFromSimplify();
  //setup the environment - load the axioms etc.
  SetupEnv();
#else
  //create the name of cvs executable file
  char *envptr = getenv("MAGICROOT");
  if(envptr == NULL) Util::Error("MAGICROOT not set ... Aborting ...\n");
  string command = string(envptr) + "/theorem-prover/simplify/Simplify";

  //create the file descriptors for communication
  int toSimplify[2],fromSimplify[2];
  if((pipe(toSimplify) == -1) || (pipe(fromSimplify) == -1)) {
    Util::Error("could not create file descriptors for communication ...\n");
  }

  //start the simplify process
  proc = fork();
  //failure
  if(proc == -1) Util::Error("could not fork for creating Simplify process ...\n");
  //child process
  if(proc == 0) {
    //close file descriptors
    if((close(toSimplify[1]) == -1) || (close(fromSimplify[0]) == -1) ||
       (close(0) == -1) || (close(1) == -1)) {
      assert(false);
    }
    //duplicate file descriptors
    if((dup(toSimplify[0]) == -1) || (dup(fromSimplify[1]) == -1)) {
      assert(false);
    }
    //exec the simplify process
    if(execl(command.c_str(),command.c_str(),"-nosc",NULL) == -1) {
      assert(false);
    }
  }
  //parent process
  else {
    Util::Message(2,"Simplify process started ...\n");
    //close file descriptors
    if((close(toSimplify[0]) == -1) || (close(fromSimplify[1]) == -1)) {
      Util::Error("could not close pipe file descriptors in parent ...\n");
    }
    //convert file descriptors into FILE*
    if(((out = fdopen(toSimplify[1],"w")) == NULL) || ((in = fdopen(fromSimplify[0],"r")) == NULL)) {
      Util::Error("could not convert file descriptors to FILE* in parent ...\n");
    }
    //set the query size and clear enum constraints
    totalQuerySize = 0;
    enumsAdded.clear();
    //read up the initial Simplify prompt
    ReadFromSimplify();
    //setup the environment - load the axioms etc.
    SetupEnv();
  }
#endif //WIN32
}

/*********************************************************************/
//shutdown the theorem prover
/*********************************************************************/
void SimplifyTP::Shutdown()
{
#ifdef WIN32
  //close process and thread handles
  if(TerminateProcess(proc.hProcess,1)) {
    CloseHandle(proc.hProcess);
    CloseHandle(proc.hThread);
    CloseHandle(in);
    CloseHandle(out);
    Util::Message(2,"Simplify process destroyed ...\n");
  } else Util::Error("could not destroy Simplify process ...\n");
#else
  //destroy the simplify process
  if(kill(proc,SIGKILL) == -1) {
    Util::Error("could not destroy Simplify process ...\n");
  } else {
    if(waitpid(proc,NULL,0) != proc) {
      Util::Error("could not free resources of Simplify process ...\n");
    } else {
      Util::Message(2,"Simplify process destroyed ...\n");
    }
  }
#endif //WIN32
}

/*********************************************************************/
//given a list of expressions create a string of their conjunction
/*********************************************************************/
string SimplifyTP::CreateConjunction(const list<const BasicExpr*> &exprs) throw (SimplifyException)
{
  //first convert all the expressions to boolean formula strings
  list<string> x;
  for(list<const BasicExpr*>::const_iterator i = exprs.begin();i != exprs.end();i++) {
    try {
      x.push_back(ConvertToBoolean(Formula(*i)));
    } catch(SimplifyException ex) {
      assert(ex == NOT_HANDLED);
      throw ex;
    }
  }
  
  //handle special cases
  if(x.size() == 0) return "TRUE";
  else if(x.size() == 1) return x.front();
  
  //create the conjunction
  string ret = "(AND";
  for(list<string>::iterator i = x.begin();i != x.end();i++) {
    ret = ret + " " + (*i);
  }
  return ret + ")";
}

/*********************************************************************/
//given a list of expressions create a string of their disjunction
/*********************************************************************/
string SimplifyTP::CreateDisjunction(const list<const BasicExpr*> &exprs) throw (SimplifyException)
{
  //first convert all the expressions to boolean formula strings
  list<string> x;
  for(list<const BasicExpr*>::const_iterator i = exprs.begin();i != exprs.end();i++) {
    try {
      x.push_back(ConvertToBoolean(Formula(*i)));
    } catch(SimplifyException ex) {
      assert(ex == NOT_HANDLED);
      throw ex;
    }
  }
  
  //handle special cases
  if(x.size() == 0) return "FALSE";
  else if(x.size() == 1) return x.front();
  
  //create the conjunction
  string ret = "(OR";
  for(list<string>::iterator i = x.begin();i != x.end();i++) {
    ret = ret + " " + (*i);
  }
  return ret + ")";
}

/*********************************************************************/
//prove that an antecedent implies a consequent. both the antecedent
//and the consequent are provided as a list of formulas. the result is
//true if the cojunction of all of the antecedents implies the
//disjunction of all of the consequents. return true if it can be
//proved and false otherwise.
/*********************************************************************/
bool SimplifyTP::ProveImplies(const list<const BasicExpr*> &ante,const list<const BasicExpr*> &cons)
{
  try {
    //create the antecedent formula
    string anteStr = CreateConjunction(ante);
    
    //create the consequent formula
    string consStr = CreateDisjunction(cons);
    
    //create the implication formula
    string formula = "(IMPLIES " + anteStr + " " + consStr + ")";
    
    //first check the cache
    int x = CheckCache(formula);
    if(x != -1) return (x == 1);
    
    //add appropriate enumeration constraints
    AddEnumConstraints(ante,cons);
    
    //send the formula to simplify
    WriteToSimplify(formula);
    totalQuerySize += formula.size();
    
    //read the reply. return true if the reply is Valid and false
    //if the reply is Invalid. flush the input.
    string reply = ReadFromSimplify();  
    if(totalQuerySize > MAX_TOTAL_QUERY_SIZE) {
      Shutdown();
      Initialise();
    }
    if(reply.find("Valid") != string::npos) {
      UpdateCache(formula,true);
      return true;
    } else if(reply.find("Invalid") != string::npos) {
      UpdateCache(formula,false);
      return false;
    } else {
      string output = "antecedent : " + Util::ExprListToString(ante) + 
	"\nconsequent : " + Util::ExprListToString(cons) + "\nformula : " + 
	formula + "\nunknown reply from Simplify : " + reply;
      Util::Error(output);
    }
    return false;
  } catch(SimplifyException ex) {
    assert(ex == NOT_HANDLED);
    return false;
  }
}
  
/*********************************************************************/
//add a constraint to the theorem prover environment. this constraint
//is provided as a C expression.
/*********************************************************************/
void SimplifyTP::AddConstraint(const BasicExpr *expr)
{
  try {
    string x = "(BG_PUSH " + Formula(expr) + ")";
    WriteToSimplify(x);
    ReadFromSimplify();
  } catch(SimplifyException ex) {
    Util::Error("unhandled exception in SimplifyTP ...\n");
  }
}

/*********************************************************************/
//read from simplfy till u get a ">"
/*********************************************************************/
string SimplifyTP::ReadFromSimplify()
{
#ifdef WIN32
  string res;
  char y[2];
  while(true) {
    DWORD bytesRead;
    if((!ReadFile(in,y,1,&bytesRead,NULL)) || (bytesRead != 1)) {
      Util::Error("error reading from Simplify ...\n");
    }
    y[1] = '\0';
    res += y;
    if(y[0] == '>') break;
  }
  return res;
#else
  string res;
  char y[2]; y[1] = '\0';
  while(true) {
    int a = fgetc(in);
    if(a == EOF) {
      Util::Error("error reading from Simplify ...\n");
    }
    y[0] = static_cast<char>(a);
    res += y;
    if(y[0] == '>') break;
  }
  return res;
#endif //WIN32
}

/*********************************************************************/
//write a line to simplify
/*********************************************************************/
void SimplifyTP::WriteToSimplify(const string &line)
{
#ifdef WIN32
  string newLine = line + "\n";
  DWORD bytesWritten;
  if((!WriteFile(out,newLine.c_str(),newLine.size(),&bytesWritten,NULL)) ||
     (bytesWritten != newLine.size())) {
    Util::Error("error writing to Simplify ...\n");
  }
#else
  fprintf(out,"%s\n",line.c_str());
  fflush(out);
#endif //WIN32
}

/*********************************************************************/
//setup the theorem proving environment. load the axioms etc.
/*********************************************************************/
void SimplifyTP::SetupEnv()
{
  //load the dereference-address axiom
  string x = "(BG_PUSH (FORALL (x) (EQ x (" + POINTER_DEREF + " (" + ADDRESS + " x)))))";
  WriteToSimplify(x);
  ReadFromSimplify();
  //load the address-dereference axiom
  string y = "(BG_PUSH (FORALL (x) (EQ x (" + ADDRESS + " (" + POINTER_DEREF + " x)))))";
  WriteToSimplify(y);
  ReadFromSimplify(); 
}

/*********************************************************************/
//convert a C expression ast to a simplify formula. throw an exception
//if the expression uses a construct that is not handled yet.
/*********************************************************************/
string SimplifyTP::Formula(const BasicExpr *node) throw (SimplifyException)
{
  //id_expression
  if(typeid(*node) == typeid(IdExpr)) {
    string x = Util::TrimString(node->ToString());
    //if the identifier begins with an underscore we rename it
    //since simplify does not allow such identifiers
    if(x.at(0) == '_') x = "alpha_renamed" + x;
    //if the identifier is a "*" introduced by a default return, we
    //change it to a fresh variable
    else if(x == "*") x = Util::NewTempVar();
    return x;
  }
  //int_const_expression
  else if(typeid(*node) == typeid(IntConstExpr)) {
    return Util::TrimString(node->ToString());
  }
  //const_expression
  else if(typeid(*node) == typeid(ConstExpr)) {
    throw NOT_HANDLED;
    //Util::Error("simplify cannot handle " + node->ToString() + "\n");
  }
  //str_expression
  else if(typeid(*node) == typeid(StrExpr)) {
    throw NOT_HANDLED;
    //Util::Error("simplify cannot handle " + node->ToString() + "\n");
  }
  //str_str_expression
  else if(typeid(*node) == typeid(StrStrExpr)) {
    throw NOT_HANDLED;
    //Util::Error("simplify cannot handle " + node->ToString() + "\n");
  }
  //stmt_expression
  else if(typeid(*node) == typeid(StmtExpr)) {
    Util::Error("simplify cannot handle " + node->ToString() + "\n");
  }
  //brack_expression
  else if(typeid(*node) == typeid(BrackExpr)) {
    const BrackExpr *x = static_cast<const BrackExpr*>(node);
    string a = Formula(x->array);
    string b = Formula(x->index);
    return "(" + POINTER_DEREF + " (+ " + a + " " + b + "))";
  }
  //par_expression
  else if(typeid(*node) == typeid(ParExpr)) {
    const ParExpr *x = static_cast<const ParExpr*>(node);
    list<BasicExpr*> &a = x->args->data;
    char buf[50];
    snprintf(buf,50,"%d ",a.size());
    string y = "(" + PROCEDURE + buf + Formula(x->proc);
    for(list<BasicExpr*>::const_iterator i = a.begin();i != a.end();i++) {
      y += " " + Formula(*i);
    }
    y += ")";
    return y;
  }
  //dot_expression
  else if(typeid(*node) == typeid(DotExpr)) {
    const DotExpr *x = static_cast<const DotExpr*>(node);
    string y = Formula(x->expr);
    string z = (x->id.at(0) == '_') ? ALPHA_RENAMED + x->id : x->id;
    return "(" + FIELD_ACCESS + " " + y + " " + z + ")";
  }
  //arrow_expression
  else if(typeid(*node) == typeid(ArrowExpr)) {
    const ArrowExpr *x = static_cast<const ArrowExpr*>(node);
    string y = Formula(x->expr);
    string z = (x->id.at(0) == '_') ? ALPHA_RENAMED + x->id : x->id;
    string a = "(" + POINTER_DEREF + " " + y + ")";
    return "(" + FIELD_ACCESS + " " + a + " " + z + ")";
  }
  //inc_expression
  else if(typeid(*node) == typeid(IncExpr)) {
    Util::Error("simplify cannot handle " + node->ToString() + "\n");
  }
  //unary_expression
  else if(typeid(*node) == typeid(UnaryExpr)) {
    const UnaryExpr *x = static_cast<const UnaryExpr*>(node);
    if(x->op == '&') {
      return "(" + ADDRESS + " " + Formula(x->expr) + ")";
    } else if(x->op == '*') {
      return "(" + POINTER_DEREF + Formula(x->expr) + ")";
    } else if(x->op == '+') {
      return "(+ 0 " + Formula(x->expr) + ")";
    } else if(x->op == '-') {
      return "(- 0 " + Formula(x->expr) + ")";
    } else if(x->op == '~') {
      return "(" + BITWISE_NOT + " " + Formula(x->expr) + ")";
    } else {
      return "(NOT " + ConvertToBoolean(Formula(x->expr)) + ")";
    }
  }
  //empty_expression
  else if(typeid(*node) == typeid(EmptyExpr)) {
    Util::Message(2,"warning : simplify cannot handle " + node->ToString() + "\n");
    throw NOT_HANDLED;
  }
  //binary_expression
  else if(typeid(*node) == typeid(BinaryExpr)) {
    const BinaryExpr *x = static_cast<const BinaryExpr*>(node);
    string y = Formula(x->lhs);
    string z = Formula(x->rhs);
 
    if(x->op == '*') return "(* " + y + " " + z + ")";
    else if(x->op == '/') Util::Error("simplify cannot handle " + node->ToString() + "\n");
    else if(x->op == '%') Util::Error("simplify cannot handle " + node->ToString() + "\n");
    else if(x->op == '+') return "(+ " + y + " " + z + ")";
    else if(x->op == '-') return "(- " + y + " " + z + ")";
    else if(x->op == MAGIC_LEFT_OP) return "(" + LEFT_SHIFT + " " + y + " " + z + ")";
    else if(x->op == MAGIC_RIGHT_OP) return "(" + RIGHT_SHIFT + " " + y + " " + z + ")";
    else if(x->op == '<') return "(< " + y + " " + z + ")";
    else if(x->op == '>') return "(> " + y + " " + z + ")";
    else if(x->op == MAGIC_LE_OP) return "(<= " + y + " " + z + ")";
    else if(x->op == MAGIC_GE_OP) return "(>= " + y + " " + z + ")";
    else if(x->op == MAGIC_EQ_OP) return "(EQ " + y + " " + z + ")";
    else if(x->op == MAGIC_NE_OP) return "(NEQ " + y + " " + z + ")";
    else if(x->op == '&') return "(" + BITWISE_AND + " " + y + " " + z + ")";
    else if(x->op == '^') return "(" + BITWISE_XOR + " " + y + " " + z + ")";
    else if(x->op == '|') return "(" + BITWISE_OR + " " + y + " " + z + ")";
    else if(x->op == MAGIC_AND_OP) {
      //convert both sub-expressions to boolean form by
      //comparing with zero
      y = ConvertToBoolean(y);
      z = ConvertToBoolean(z);
    
      return "(AND " + y + " " + z + ")";
    } else if(x->op == MAGIC_OR_OP) {
      //convert both sub-expressions to boolean form by
      //comparing with zero
      y = ConvertToBoolean(y);
      z = ConvertToBoolean(z);
    
      return "(OR " + y + " " + z + ")";
    } else Util::Error("simplify cannot handle " + node->ToString() + "\n");
  }
  //quest_expression
  else if(typeid(*node) == typeid(QuestExpr)) {
    const QuestExpr *x = static_cast<const QuestExpr*>(node);
    //convert an expression of the form A ? B : C to (A && B)
    //|| (!A && C)
    BasicExpr *a = new BinaryExpr(x->cond,x->tcase,MAGIC_AND_OP);
    BasicExpr *b = x->cond->Negate();
    BasicExpr *c = new BinaryExpr(b,x->ecase,MAGIC_AND_OP);
    delete b;
    BasicExpr *d = new BinaryExpr(a,c,MAGIC_OR_OP);
    delete a; delete c;
    string res = Formula(d);
    delete d;
    return res;
  }
  //assign_expression
  else if(typeid(*node) == typeid(AssignExpr)) {
    const AssignExpr *x = static_cast<const AssignExpr*>(node);
    return Formula(x->lhs);
  }
  //illegal expression
  else {
    Util::Error("simplify cannot handle " + node->ToString() + "\n");
  }
}

/*********************************************************************/
//convert a formula to boolean form by comparing with 0
/*********************************************************************/
string SimplifyTP::ConvertToBoolean(const string &old)
{
  if((old.at(0) != '(') || (old.find("(+") == 0) ||
     (old.find("(-") == 0) || (old.find("(*") == 0) ||
     (old.find("(" + FIELD_ACCESS) == 0) ||
     (old.find("(" + POINTER_DEREF) == 0) ||
     (old.find("(" + ADDRESS) == 0) ||
     (old.find("(" + BITWISE_AND) == 0) ||
     (old.find("(" + BITWISE_XOR) == 0) ||
     (old.find("(" + BITWISE_OR) == 0) ||
     (old.find("(" + LEFT_SHIFT) == 0) ||
     (old.find("(" + RIGHT_SHIFT) == 0) ||
     (old.find("(" + PROCEDURE) == 0) ||
     (old.find(PROCEDURE) == 0)) {
    return "(NEQ " + old + " 0)";
  } else return old;
}

/*********************************************************************/
//check if a formula belongs to a cache. returns 1 if the formula is
//in the true cache, 0 if it is in the false cache and -1
//otherwise. also updates the theorem prover statistics accordingly.
/*********************************************************************/
int SimplifyTP::CheckCache(const string &formula)
{
  if(trueCache.count(formula) != 0) {
    Statistics::tpInCache++;
    return 1;
  } else if(falseCache.count(formula) != 0) {
    Statistics::tpInCache++;
    return 0;
  } else {
    Statistics::tpCalls++;
    Statistics::tpCacheSize++;
    return -1;
  }
}

/*********************************************************************/
//add appropriate enumeration constraints given two lists of
//formulas
/*********************************************************************/
void SimplifyTP::AddEnumConstraints(const list<const BasicExpr*> &ante,const list<const BasicExpr*> &cons)
{  
  //combine the two lists
  list<const BasicExpr*> all = ante;
  all.insert(all.end(),cons.begin(),cons.end());
  
  //compute the set of ids in the expressions
  set<string> workList;
  for(list<const BasicExpr*>::const_iterator i = all.begin();i != all.end();i++) {
    set<string> ids = Util::ComputeLvalues(*i);
    for(set<string>::const_iterator j = ids.begin();j != ids.end();j++) {
      if(enumsAdded.count(*j) == 0) {
	workList.insert(*j);
      }
    }
  }
  
  //get the enum map
  const map<string,const BasicExpr*> &enums = Database::enums;
  
  //now do the stuff
  while(!workList.empty()) {
    string a = *(workList.begin());
    workList.erase(workList.begin());
    if(enums.count(a) != 0) {
      if(enumsAdded.count(a) == 0) {
	const BasicExpr *lhs = new IdExpr(a);
	const BasicExpr *rhs = enums.find(a)->second;
	BasicExpr *eq = new BinaryExpr(lhs,rhs,MAGIC_EQ_OP);
	AddConstraint(eq);
	enumsAdded.insert(a);
	
	set<string> b = Util::ComputeLvalues(rhs);
	for(set<string>::const_iterator i = b.begin();i != b.end();i++) {
	  if(enumsAdded.count(*i) == 0) {
	    workList.insert(*i);
	  }
	}
	
	delete lhs; delete eq;
      }
    }
  }
}

/*********************************************************************/
//add a new cache entry. the first argument is the entry. the second
//argument is the type of the cache to be updated.
/*********************************************************************/
void SimplifyTP::UpdateCache(const string &formula,bool type)
{
  if(Database::THEOREM_PROVER_CACHE) {
    if(type) {
      if(static_cast<int>(trueCache.size()) > Database::THEOREM_PROVER_CACHE_SIZE) {
	trueCache.clear();
      }
      trueCache.insert(formula);
    } else {
      if(static_cast<int>(falseCache.size()) > Database::THEOREM_PROVER_CACHE_SIZE) {
	falseCache.clear();
      }
      falseCache.insert(formula);
    }
  }
}

/*********************************************************************/
//end of SimplifyTP.cpp
/*********************************************************************/
